"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""


# ###################################
# # MMA plugin utilities            #
# #                                 #
# # written by Ignazio Di Napoli    #
# # <neclepsio@gmail.com>           #
# ###################################

import textwrap, traceback, sys, re
from collections import defaultdict

from mma.MMA.common import error, warning
from mma.MMA.macro import macros
import mma.MMA.gbl as gbl

from mma.MMA.termsize import getTerminalSize
termwidth = getTerminalSize()[0]-1

try:
    basestring
except NameError:
    basestring = str
  

# ###################################
# # Plugin configuration            #
# ###################################

class _pluginInfo (object):
    def __init__(self):
        self.NAME = None
        self.DESCRIPTION = ""
        self.SYNOPSIS = ""
        self.AUTHOR = ""
        self.TRACKTYPES = None
        self.ARGUMENTS = []
        self.DOC = ""
        self.COMMANDS = []
    
_plugins = defaultdict(_pluginInfo)

def _getCallerFileName():
    stackFileNames = [s[0] for s in traceback.extract_stack()]
    myName = stackFileNames[-1]
    for name in stackFileNames[::-1]:
        if name != myName:
            return name
    return None

    
def _getCallerModule():
    f = _getCallerFileName()
    if f is None:
        return None
        
    for module in sys.modules.values():
        if module is None or not hasattr(module, "__file__"):
            continue
        if module.__file__.rpartition(".")[0] != f.rpartition(".")[0]:
            continue
        return module
    
    return None
    
    
def _P():
    f = _getCallerFileName()
    
    # plugInName is not available at module initialization, but is called
    # afterwards. So, we assign it when called later.
    if _plugins[f].NAME is None:
        module = _getCallerModule()
        if module is not None and hasattr(module, "plugInName"):
            _plugins[f].NAME = "@" + module.plugInName['name']

    return _plugins[f]


# A short description.
def setDescription(descr):
    _P().DESCRIPTION = descr

# Synopsis.
def setSynopsis(syno):
    _P().SYNOPSIS = syno

# Author.
def setAuthor(author):
    _P().AUTHOR = author

# Track types to which the plugin applies; if empty everything is accepted.
# If None, this is a non-track plugin.
def setTrackType(*types):
    _P().TRACKTYPES = types[:]

# A list of arguments for the plugin; each is specified with [name, default, doc]. 
# If no default is provided, use None to trigger an error if the user not specifies it.
def addArgument(name, value, doc):
    _P().ARGUMENTS.append((name, value, doc))

# Simple documentation. %NAME% is replaced by plugin name.
def setPluginDoc(doc):
    _P().DOC = doc.strip()
    
# Minimum MMA version required.
def setMinMMAVersion(major, minor):
    vre = re.compile("([0-9]+)\\.([0-9]+)\\.?(.*)")
    m = vre.match(gbl.version)
    if m is None:
        error("Plugin utils: version %s not recognized." % gbl.version)
    cur_major = int(m.group(1))    
    cur_minor = int(m.group(2))    
    cur_rev = m.group(3)
    
    if cur_major > major:
        return
    if cur_major == major and cur_minor >= minor:
        return
    
    error("Plugin requires MMA version {:02}.{:02} or newer.".format(major, minor))

   
# Returns plugin name
def getName():
    return _P().NAME
    
     
    
# ###################################
# # Utility functions               #
# ###################################

# If you don't send all the commands together the result is commands 
# are reversed since each is pushed as the very next command to be executed.
# So we save all the commands (with addCommand) and send them at the end
# (with sendCommands).

def addCommand(strings):
    if isinstance(strings, basestring):
        strings = [strings]
    _P().COMMANDS.extend(strings)

    
def sendCommands():
    # All values have to be lists of words. Not a string per line!
    ret = [l.split() for l in _P().COMMANDS]
    _P().COMMANDS = []
    
    #print (ret)

    # The next line does the input stream push. Note that we are using
    # the current line number of the file for each line.
    gbl.inpath.push(ret, [gbl.lineno] * len(ret))

    
def parseCommandLine(line, allowUnknown=False):
    s = " ".join(line)
    subst = [
        [" =", "="],
        ["= ", "="],
        ["  ", " "],
        [", ", ","],
        [" ,", ","],
    ]
    for s1, s2 in subst:
        while s1 in s:
            s = s.replace(s1, s2)
    
    res = {name:default for name, default, _ in _P().ARGUMENTS}
    positional = True
    for name, default, _ in _P().ARGUMENTS:
        res[name] = default
        
    for i, value in enumerate(s.split(",")):
        if value == "":
            continue
            
        if "=" in value:
            positional = False
            n, _, v = value.partition("=")
            if n not in res: 
                if not allowUnknown:
                    error("Plugin {}: unexpected argument name {}.".format(_P().NAME, n))
                else:
                    n = n.upper()
                    
        elif positional:
            if i >= len(_P().ARGUMENTS):
                error("Plugin {}: unexpected argument provided ({}).".format(_P().NAME, value))
            n, _, _ = _P().ARGUMENTS[i]
            v = value
            
        else:
            error("Plugin {}: positional argument after named argument is not allowed.".format(_P().NAME))
            
        res[n] = v
    
    for k, v in res.items():
        if v is None:
            printUsage()
            error("Plugin {}: no value for argument {}.".format(_P().NAME, k))
    
    return res
    

def _printUsage(plugin):
    width = termwidth
    
    lines = []
    lines.append((0, "Plugin {}".format(plugin.NAME)))
    if plugin.AUTHOR != "":
        lines.append((0, plugin.AUTHOR))
    lines.append((0, ""))
    if plugin.DESCRIPTION != "":
        lines.append((0, plugin.DESCRIPTION))
        lines.append((0, ""))
    lines.append((0, "SYNOPSIS"))
    if plugin.SYNOPSIS != "":
        for line in plugin.SYNOPSIS.splitlines():
            lines.append((0, line))
        lines.append((0, ""))
    else:
        t = ""
        if plugin.TRACKTYPES is not None:
            t += "Track "
        t += plugin.NAME + " "

        for name, _, _ in plugin.ARGUMENTS:
            t += name + ", " 

        lines.append(([4, 8], t.strip().rstrip(",")))
        lines.append((0, ""))

    if plugin.TRACKTYPES is not None and len(plugin.TRACKTYPES) > 0:
        if len(plugin.TRACKTYPES) == 1:
            lines.append((4, "Track type must be {}.".format(plugin.TRACKTYPES[0])))
        else:
            lines.append((4, "Track type must be one of {}.".format(", ".join(plugin.TRACKTYPES[:-1]) + " or " + plugin.TRACKTYPES[-1])))
        lines.append((0, ""))

    if len(plugin.ARGUMENTS) > 0:
        lines.append((0, "ARGUMENTS"))
        lines.append((0, ""))

        for name, default, doc in plugin.ARGUMENTS:
            defvalue = (" (default: "+default+")" if default is not None else "")
            lines.append((4, name + defvalue))
            lines.append((8, doc))
            lines.append((0, ""))

        lines.append((4, "You can also use name=value syntax, like Call (see MMA documents)."))
        lines.append((0, ""))
        
    lines.append((0, "DESCRIPTION"))
    lines.append((0, ""))
    for l in plugin.DOC.replace("\r", "").replace("%NAME%", plugin.NAME).split("\n"):
        lines.append((4, l))
    
    for indent, line in lines:
        if line.strip() == "":
            print 
        else:
            if isinstance(indent, list):
                firstindent, indent = indent
            else:
                firstindent = indent
            
            llines = textwrap.wrap(line, width-indent)
            for i, lline in enumerate(llines):
                if i == 0 and firstindent != indent:
                    print(" " * firstindent + lline + (" \\" if len(llines) > 1 else ""))
                else:
                    print(" " * indent + lline)

                    
def printUsage():
    _printUsage(_P())
       
    
def checkTrackType(name):
    if _P().TRACKTYPES is None or len(_P().TRACKTYPES) == 0:
        return

    for t in _P().TRACKTYPES:
        if name.upper() == t.upper() or name.upper().startswith(t.upper() + "-"):
            return
    
    if len(_P().TRACKTYPES) == 1:
        error("Plugin {}: track type must be {}.".format(_P().NAME, _P().TRACKTYPES[0]))
    else:
        error("Plugin {}: track type must be one of {}.".format(_P().NAME, ", ".join(_P().TRACKTYPES[:-1]) + " or " + _P().TRACKTYPES[-1]))


def getTrackType(name):
    if "-" in name:
        name = name.split("-")[0]
    return name.upper()
            
error = error
warning = warning

def getVar(name):
    return macros.vars[name]
       
def setVar(name,value):
    macros.vars[name.upper()] = str(value)
       
def getSysVar(name):
    return macros.sysvar(name)
    
    



# patBass.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""


import mma.MMA.notelen
import mma.MMA.harmony
import mma.MMA.ornament

from . import gbl
from mma.MMA.common import *
from mma.MMA.pat import PC, Pgroup


class Bass(PC):
    """ Pattern class for a bass track. """

    vtype = 'BASS'

    def getPgroup(self, ev):
        """ Get group for bass pattern.

            Fields - start, length, note, volume

        """

        if len(ev) != 4:
            error("There must be n groups of 4 in a pattern definition, "
                  "not <%s>" % ' '.join(ev))

        a = Pgroup()

        a.offset = self.setBarOffset(ev[0])
        a.duration = mma.MMA.notelen.getNoteLen(ev[1])

        offset = ev[2]
        n = offset[0]
        if n in "1234567":
            a.noteoffset = int(n) - 1
        else:
            error("Note offset in Bass must be '1'...'7', not '%s'" % n)

        n = offset[1:2]
        if n == "#":
            a.accidental = 1
            ptr = 2
        elif n == 'B' or n == '&':
            a.accidental = -1
            ptr = 2
        else:
            a.accidental = 0
            ptr = 1

        a.addoctave = 0

        for n in ev[2][ptr:]:
            if n == '+':
                a.addoctave += 12
            elif n == '-':
                a.addoctave -= 12
            else:
                error("Only '- + # b &' are permitted after a noteoffset, not '%s'" % n)

        a.vol = stoi(ev[3], "Note volume in Bass definition not int")

        return a

    def restart(self):
        self.ssvoice = -1

    def trackBar(self, pattern, ctable):
        """ Do the bass bar.

        Called from self.bar()

        """

        sc = self.seq

        for p in pattern:
            ct = self.getChordInPos(p.offset, ctable)
            
            if ct.bassZ:
                continue

            note = ct.chord.scaleList[p.noteoffset] + p.addoctave + p.accidental

            if not self.harmonyOnly[sc]:
                notelist = [(note, p.vol)]
            else:
                notelist = []

            if self.harmony[sc]:
                h = mma.MMA.harmony.harmonize(self.harmony[sc], note, ct.chord.noteList)
                harmlist = list(zip(h, [p.vol * self.harmonyVolume[sc]] * len(h)))
            else:
                harmlist = []

            offset = p.offset
            if self.ornaments['type']:
                offset = mma.MMA.ornament.doOrnament(self, notelist,
                                        self.getChordInPos(offset, ctable).chord.scaleList, p)
                notelist = []

            self.sendChord(notelist + harmlist, p.duration, offset)




# mdefine.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

This class is used to parse lines of MDEFINE and stores
the sequences for later recall.

"""

import mma.MMA.midiC
import mma.MMA.midiM

from . import gbl
from   mma.MMA.common import *


def mdefine(ln):
    """ Set a midi seq pattern. """

    if not ln:
        error("MDefine needs arguments")

    name = ln[0]
    if name.startswith('_'):
        error("Names with a leading underscore are reserved")

    if name.upper() == 'Z':
        error("The name 'Z' is reserved")

    mdef.create(name, ' '.join(ln[1:]))


def trackMdefine(name, ln):
    """ Set a midi seq pattern. Ignore track name."""

    mdefine(ln)


class Mdefine:

    def __init__(self):
        self.defs = {}

    def get(self, name):
        """ Return a predefined MIDI pattern."""

        try:
            return self.defs[name]
        except:
            error("The MDEFINE pattern %s has not been defined" % name)

    def create(self, name, ln):
        """ Parse a MDEFINE line.

            The line must be in the form:

                NAME <beat> <ctrl> <dat> [; ...]

        """

        name = name.upper()

        ln = ln.rstrip('; ')     # dump trailing ';' and whitespace
        ln = ln.split(';')
        evs = []
        for l in ln:
            l = l.split()

            if len(l) == 1:
                evs.extend(self.get(l[0].upper()))
                continue

            if len(l) != 3:
                error("MDEFINE sequence must have 3 values: Beat, Ctrl, Datum")

            off = stof(l[0], "Value for offset must be integer/float")

            c = mma.MMA.midiC.ctrlToValue(l[1])
            if c < 0:
                c = stoi(l[1])
                if c < 0 or c > 0x7f:
                    error("Controller values must be 0x00 to 0x7f")

            d = stoi(l[2])
            if d < 0 or d > 0x7f:
                error("MIDI Control Datum value must be 0x00 to 0x7f")

            evs.append([off, mma.MMA.midiM.packBytes(c, d)])

        self.defs[name] = evs

mdef = Mdefine()



# roman.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

Roman numeral chord to standard notations.
"""

from mma.MMA.common import *
from mma.MMA.keysig import keySig


# Table of scales ... a list of 7 notes for each possible Major/Minor scale

majTable = {'C': ('C', 'D', 'E', 'F', 'G', 'A', 'B'),
            'C#': ('C#', 'D#', 'E#', 'F#', 'G#', 'A#', 'C'),
            'D': ('D', 'E', 'F#', 'G', 'A', 'B', 'C#'),
            'Db': ('Db', 'Eb', 'F', 'Gb', 'Ab', 'Bb', 'C'),
            'D#': ('D#', 'F', 'G', 'G#', 'A#', 'C', 'D'),
            'Eb': ('Eb', 'F', 'G', 'Ab', 'Bb', 'C', 'D'),
            'E': ('E', 'F#', 'G#', 'A', 'B', 'C#', 'D#'),
            'F': ('F', 'G', 'A', 'Bb', 'C', 'D', 'E'),
            'F#': ('F#', 'G#', 'A#', 'B', 'C#', 'D#', 'F'),
            'Gb': ('Gb', 'Ab', 'Bb', 'B', 'Db', 'Eb', 'F'),
            'G': ('G', 'A', 'B', 'C', 'D', 'E', 'F#'),
            'G#': ('G#', 'A#', 'C', 'C#', 'D#', 'F', 'G'),
            'Ab': ('Ab', 'Bb', 'C', 'Db', 'Eb', 'F', 'G'),
            'A': ('A', 'B', 'C#', 'D', 'E', 'F#', 'G#'),
            'A#': ('A#', 'C', 'D', 'D#', 'F', 'G', 'A'),
            'Bb': ('Bb', 'C', 'D', 'Eb', 'F', 'G', 'A'),
            'B': ('B', 'C#', 'D#', 'E', 'F#', 'G#', 'A#')}

minTable = {'C': ('C', 'D', 'Eb', 'F', 'G', 'Ab', 'Bb'),
            'C#': ('C#', 'D#', 'E', 'F#', 'G#', 'A', 'B'),
            'Db': ('Db', 'Eb', 'E', 'Gb', 'Ab', 'A', 'B'),
            'D': ('D', 'E', 'F', 'G', 'A', 'Bb', 'C'),
            'D#': ('D#', 'F', 'F#', 'G#', 'A#', 'B', 'C#'),
            'Eb': ('Eb', 'F', 'Gb', 'Ab', 'Bb', 'B', 'Db'),
            'E': ('E', 'F#', 'G', 'A', 'B', 'C', 'D'),
            'F': ('F', 'G', 'Ab', 'Bb', 'C', 'Db', 'Eb'),
            'F#': ('F#', 'G#', 'A', 'B', 'C#', 'D', 'E'),
            'Gb': ('Gb', 'Ab', 'A', 'B', 'Db', 'D', 'E'),
            'G': ('G', 'A', 'A#', 'C', 'D', 'D#', 'F'),
            'G#': ('G#', 'A#', 'B', 'C#', 'D#', 'E', 'F#'),
            'Ab': ('Ab', 'Bb', 'B', 'Db', 'Eb', 'E', 'Gb'),
            'A': ('A', 'B', 'C', 'D', 'E', 'F', 'G'),
            'A#': ('A#', 'C', 'C#', 'D#', 'F', 'F#', 'G#'),
            'Bb': ('Bb', 'C', 'Db', 'Eb', 'F', 'Gb', 'Ab'),
            'B': ('B', 'C#', 'D', 'E', 'F#', 'G', 'A')}

uroman = {'I': 0, 'II': 1, 'III': 2, 'IV': 3, 'V': 4, 'VI': 5, 'VII': 6}
lroman = {'i': 0, 'ii': 1, 'iii': 2, 'iv': 3, 'v': 4, 'vi': 5, 'vii': 6}
arabic = {'1': 0, '2': 1, '3': 2, '4': 3, '5': 4, '6': 5, '7': 6}

doubleflat = {'Cbb': 'Bb', 'Dbb': 'C', 'Ebb': 'D', 'Fbb': 'Eb',
              'Gbb': 'F', 'Abb': 'G', 'Bbb': 'A'}

doublesharp = {'C##': 'D', 'D##': 'E', 'E##': 'F#', 'F##': 'G',
               '#': 'A', 'A##': 'B', 'B##': 'C#'}

convertable = {'m' + chr(176): 'dim3', 'm0': 'dim3', 'mo': 'dim3', 'mO': 'dim3',
               'm' + chr(176) + '7': 'dim7', 'm07': 'dim7', 'mo7': 'dim7', 'mO7': 'dim7',
               'm' + chr(248) + '7': 'm7b5', 'm-07': 'm7b5', 'm-o7': 'm7b5', 'm-O7': 'm7b5'}


def rvalue(s):
    """ Convert a roman or arabic numeral to value (-1). """

    if s in uroman:
        return uroman[s]
    elif s in lroman:
        return lroman[s]
    elif s in arabic:
        return arabic[s]
    else:
        if s[0].isdigit:
            error("Unknown Arabic value '%s'. Use 1 to 7." % s)
        else:
            error("Unknown Roman numeral '%s'. Use 'I' to 'VII' in all u/l case." % s)


def convert(sym):
    """ Convert a roman numeral to a standard chord name. """

    keysig, minor = keySig.kName

    # figure number of roman numerals leading symbol

    sym = list(sym)
    rm = ''
    while(sym) and sym[0] in ('I', 'V', 'i', 'v'):
        rm += sym.pop(0)
    sym = ''.join(sym)

    if rm[0].islower():
        isminor = True
    else:
        isminor = False
    offset = rvalue(rm)

    # convert the roman to a pitch ... just a table lookup

    if minor:
        pitch = minTable[keysig][offset]
    else:
        pitch = majTable[keysig][offset]

    """
        Adjust the pitch if the remainder starts with a #  or b. (Note, '&'
        was converted to 'b' early in the chord parser.

        This permits technically incorrect things like 'Ib' which end up (in C)
        as 'Cb'. Useful with dim chords. We also need to worry about doubles!
    """

    if sym.startswith('b') or sym.startswith('#'):
        pitch += sym[0]
        sym = sym[1:]

        if pitch.endswith('#b') or pitch.endswith('b#'):  # 'b#' cancel each other
            pitch = pitch[:-2]

        elif pitch.endswith('##'):
            pitch = doublesharp[pitch]

        elif pitch.endswith('bb'):
            pitch = doubleflat[pitch]

    """ Now translate the quality. This is whatever was left after
        stripping off the number and sharp/flat. Two uglies:
          - some names are different in RN and standard, ie 07 .. dim7
          - lowercase == minor, so we add in 'm' to start of the
            leftover quality, unless it's there already. "v07" becomes "Xdim7",
            and "Vm0" (wrong!) still works and becomes Xdim3"
          - special trap for double 'm's. Hide conversion of "vm7" which becomes
            "Xmm7" and then "Xm7".
    """

    if isminor and not sym.startswith('m'):
        sym = 'm' + sym

    if sym in convertable:
        sym = convertable[sym]

    return pitch + sym



#!/usr/bin/python 

"""
The program "MMA - Musical Midi Accompaniment" and the associated
modules distributed with it are protected by copyright.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import sys
import os
import platform


def cli():
    # Ensure a proper version is available.
    # This test forces 2.6+ or 3.x
    if sys.version_info[0] == 2 and sys.version_info[1] < 6:
        print ("\nYou need a more current version of Python to run mma.MMA.")
        print ("We're looking for something equal or greater than version 2.6")
        print ("Current Python version is %s\n" % sys.version)
        sys.exit(0)


    """ MMA uses a number of application specific modules. These should
        be installed in a mma modules directory or in your python
        site-packages directory). MMA searches for the modules
        directory and pre-pends the first found to the python system list.

        If you end up with mma running (ie: it finds its modules), but
        can't find libraries you then probably have the modules installed
        in python-site and lib, include, etc. are NOT in one of the following
        locations (lib and include must be in the same directory).

        Note: it is quite possible with this method to have the modules
        somewhere in python-site (NOT supported by MMA's standard installs)
        and to have MMA's libraries in a different location (ie: /usr/share/mma).
    """

    platform_so = platform.system()

    if platform_so == 'Windows':
        dirlist = ( sys.path[0], "c:/mma", "c:/program files/mma", ".")
    elif platform_so == 'Darwin':
        dirlist = ( sys.path[0], "/Users/Shared/mma", 
                "/usr/local/share/mma", "/usr/share/mma", '.' )
    else:
        dirlist = ( sys.path[0], "/usr/local/share/mma", "/usr/share/mma", '.' )

    for d in dirlist:
        moddir = os.path.join(d, 'MMA')
        if os.path.isdir(moddir):
            if not d in sys.path:
                sys.path.insert(0, d)
            MMAdir = d
            break

    # Call the mainline code. Hopefully, byte-compiled.
    # NOTE: the variables MMAdir and platform are read (imported)
    #       by gbl.py.
    import mma.MMA.main


# rpitch.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>


"""

from mma.MMA.common import *
from . import gbl
import mma.MMA.debug 

import random

class Rpitch:
    def __init__(self):
        self.scale = 'CHROMATIC'
        self.rate = .25   # 25%, apply to all notes
        self.offsets = []
        self.bars = []

def getOpts(self):
    """ Return string of current settings. """

    r = self.rPitch

    if not r:
        return "None"

    scale = r.scale
    rate = int(r.rate * 100)
    bars = ','.join([str(a + 1) for a in r.bars])

    offsets = r.offsets
    if not offsets:
        offsets = 'None'
    else:
        offsets = ','.join([str(a) for a in offsets])

    return "Rate=%s ScaleType=%s Offsets=%s Bars=%s" % (rate, scale, offsets, bars)

def setRPitch(name, ln):
    """ Set the pitch modifer.

        options:    Scale or ScaleType - chromatic, scale, chord
                    Rate - a percentage to apply. Default = 100%
                    Offsets - value to use. The values are offsets
                       and are applied to the current note.
    """

    self = gbl.tnames[name]

    msg = "%s RPitch" % self.name

    if self.vtype == 'DRUM':
        error("RPitch cannot be appied to a drum track %s." % self.name)

    if not ln or len(ln) == 1 and ln[0].upper() in ("NONE", "OFF"):
        self.rPitch = None

    else:
        ln, opts = opt2pair(ln, toupper=True)

        if ln:
            error("%s all settings must be Cmd=Opt pairs, not '%s'" % (msg, ln))

        self.rPitch = rp = Rpitch()

        for o,c in opts:
            if o == 'SCALETYPE' or o == 'SCALE':
                if not c in ('CHROMATIC', 'SCALE', 'CHORD'):
                    error("%s Rpitch ScaleType: Only Chromatic, Scale and Chord "
                      "are valid." % self.name)
                rp.scale = c
                
            elif o == 'BARS':
                l = []
                if c:
                    for a in c.split(','):
                        a = stoi(a)
                        if a < 0:
                            error("%s Bars must be positive values, not '%s'." % (msg, a))
                        if a < 1 or a > gbl.seqSize:
                            warning("%s Bars: setting of %s may be ignored since "
                            "SeqSize is %s." % (msg, a, gbl.seqSize))
                        l.append(a - 1)
                rp.bars = l[:]

            elif o == 'RATE':
                v = stoi(c)
                if v < 0:
                    error("% must be positive, not '%s'." % (msg, v))
                rp.rate = v / 100.

            elif o == 'OFFSETS':
                l = []
                if c.upper() == 'NONE' or c == '0':
                    l = []
                else:
                    for a in c.split(','):
                        if a.startswith('-'):
                            a = a[1:]
                            neg = -1
                        else:
                            neg = 1
                        if '-' in a:
                            a,b=a.split('-',1)
                            a = stoi(a) * neg
                            b = stoi(b)
                            if b < a:
                                a,b=b,a
                            l.extend(range(a,b+1))
                        else:
                            a = stoi(a) * neg
                            l.append(a)
                for a in l:
                    if a < -12 or a > 12:
                        warning("%s value of %s is large." % (msg, a))
                        break
                rp.offsets = l[:]

            else:
                error("%s RPitch %s Unknown options." % o)


    if mma.MMA.debug.debug:
         mma.MMA.debug.trackSet(self.name, "RPitch")

def doRpitch(self, note):
    """ Apply rpitch setting to note. Returns modified note."""

    mode = self.rPitch.scale
    ch = self.rPitch.offsets

    if random.random() > self.rPitch.rate or not ch:
        return note

    # we have a bar list but the current bar isn't there, just return.
    if self.rPitch.bars and gbl.seqCount not in self.rPitch.bars:
        return note

    if mode == 'CHROMATIC':
        note += random.choice(ch)

    else:
        notelist = None
        if mode == 'SCALE' and self.currentChord and self.currentChord.chord.scaleList:
            notelist = list(self.currentChord.chord.scaleList)
        elif mode == 'CHORD' and self.currentChord and self.currentChord.chord.noteList:
            notelist = list(self.currentChord.chord.noteList)

        if notelist:

            # select the offset to use from the offsets list
            change = random.choice(ch)
            if change < 0:
                neg = -1
                change = abs(change)
            else:
                neg = 1

            # Make sure scale list is long enuf. to select from
            # Simple: just keep doubling up the list. Note we need
            # to add the octave adjustment each time, and to keep incrementing it!
            o = 12
            while len(notelist) <= change:
                notelist.extend( [ i + o for i in notelist])
                o += 12
            note += ( notelist[change] * neg )

    # default fallthough
    while note > 127:
        note -= 12
    while note < 0:
        note += 12
    return note



# file.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import sys
import os
from . import gbl
from   mma.MMA.common import *
import mma.MMA.debug

PY3 = sys.version_info[0] == 3


def fixfname(f):
    """ Convert embedded space characters in filename to real spaces.

        Originally this was done with .decode("string-escape") but that
        doesn't work with windows path names. So, now we just replace
        any \x20 sequences with single spaces.
    """

    f = f.replace('\\x20', ' ')
    return os.path.expanduser(f)


def locFile(name, lib):
    """ Locate a filename.

        This checks, in order:
          lib/name + .mma
          lib/name
          name + .mma
          name
    """

    ext = gbl.EXT
    exists = os.path.exists

    name = fixfname(name)  # for ~ expansion only
    
    if lib:
        if not name.endswith(ext):
            t = os.path.join(lib, name + ext)
            if exists(t):
                return t
        t = os.path.join(lib, name)
        if exists(t):
            return t

    if not name.endswith(ext):
        t = name + ext
        if exists(t):
            return t

    if exists(name):
        return name

    return None


###########################
# File read class
###########################


class ReadFile:

    class FileData:
        """ After reading the file in bulk it is parsed and stored in this
            data structure. Blanks lines and comments are removed.
        """

        def __init__(self, lnum, data, label):
            self.lnum = lnum
            self.data = data
            self.label = label

    def __init__(self, fname):

        self.fdata = fdata = []
        self.lastline = None
        self.lineptr = None
        self.fname = None

        self.que = []     # que for pushed lines (mainly for REPEAT)
        self.qnums = []
        #self.atEOFlines = []

        dataStore = self.FileData  # shortcut to avoid '.'s

        if fname == 1:
            inpath = sys.stdin
        else:
            try:
                if PY3:
                    inpath = open(fname, 'r', encoding=gbl.encoding)
                else:
                    inpath = open(fname, 'r')
            except IOError:
                error("Unable to open '%s' for input" % fname)
 
        if mma.MMA.debug.debug or mma.MMA.debug.showFilenames:
            dPrint("Opening file '%s'." % fname)

        self.fname = fname

        """ Read entire file, line by line:
             - strip off blanks, comments
             - join continuation lines
             - parse out LABELS
             - create line numbers
        """

        lcount = 0
        label = ''
        labs = []     # track label defs, error if duplicate in same file
        nlabs = []    # track linenumber label defs
        inComment = False  # multiline comment flag

        while 1:
            l = inpath.readline()

            if not l:        # EOF
                if inComment:
                    error("Multiline comment (/* */) not terminated.")
                break

            l = l.strip()
            lcount += 1

            if not l:
                continue

            # join lines ending in '\' - the strip() makes this the last char

            while l[-1] == '\\':
                l = l[0:-1] + ' ' + inpath.readline().strip()
                lcount += 1

            """ input cleanup ... for now the only cleanup is to convert
                0xa0 (non-breakable space) to 0x20 (regular space).
            """

            l = l.replace('\xa0', '\x20')

            # multiline comments
            if inComment:
                if "/*" in l:
                    warning ("Block comment start '/*' found inside comment.")
                if "*/" in l:
                    _, _, l = l.partition("*/")
                    inComment = False
                else:
                    continue

            l = l.split('//', 1)[0]  # Strip off line comment

            while "/*" in l:
                if "*/" in l:
                    t1, _, t2 = l.partition("/*")
                    t3, _, t4 = t2.partition("*/")
                    l = t1 + t4
                else:
                    l = l.split("/*")[0]
                    inComment = True

            """ Splits the remaining line into tokens using whitespace delimiters.
                Note that split() will strip off trailing and leading
                spaces, so a strip() is not needed here.
            """

            l = l.split()

            if not l:
                continue

            """ Parse out label lines. There are 2 different kinds of labels:
                - LABEL XXX
                   and
                - NNN

                The first kind is treated as an exclusive. If a NNN label or previous
                XXX duplicates, an error is generated.

                The LINE NUMBER type is not exclusive. If a duplicate NNN is found, the
                last one is used.

                XXX NNN types can not duplicate each other.

                Also note that XXX lines are stripped from input as well as NNN lines
                with only a NNN.
            """

            if l[0].upper() == 'LABEL':
                if len(l) != 2:
                    gbl.lineno = lcount
                    error("Usage: LABEL <string>")
                label = l[1].upper()
                if label[0] == '$':
                    gbl.lineno = lcount
                    error("Variables are not permitted as labels")
                if label in labs:
                    gbl.lineno = lcount
                    error("Duplicate label specified in line %s" % lcount)
                elif label in nlabs:
                    gbl.lineno = lcount
                    error("Label '%s' duplicates line number label" % label)
                labs.append(label)

            elif l[0].isdigit():
                label = l[0]

                if label in labs:
                    gbl.lineno = lcount
                    error("Line number '%s' duplicates LABEL" % label)

                if not label in nlabs:
                    nlabs.append(label)
                else:
                    for i, a in enumerate(fdata):
                        if a.label == label:
                            fdata[i].label = ''

            else:
                label = None

            # Save the line, linenumber and (maybe) the label.

            fdata.append(dataStore(lcount, l, label))

        inpath.close()

        self.lineptr = 0
        self.lastline = len(fdata)

    def toEof(self):
        """ Move pointer to End of File. """

        self.lineptr = self.lastline+1
        self.que = []
        self.qnums = []

    def goto(self, l):
        """ Do a goto jump.

            This isn't perfect, but is probably the way most GOTOs work. If
            inside a repeat/if then nothing more is processed. The jump is
            immediate. Of course, you'll run into problems with missing
            repeat/repeatend if you try it. Since all repeats are stacked
            back into the que, we just delete the que. Then we look for a
            matching label in the file line array.

            Label search is linear. Not too efficient, but the lists
            will probably never be that long either.

        """

        if not l:
            error("No label specified")

        if self.que:
            self.que = []

        for i, a in enumerate(self.fdata):
            if a.label == l:
                self.lineptr = i
                return

        error("Label '%s' has not been set" % l)

    def pushEOFline(self, ln):
        self.fdata.append(self.FileData(gbl.lineno, ln, ''))
        self.lastline+=1

    def push(self, q, nums):
        """ Push a list of lines back into the input stream.

            Note: This is a list of semi-processed lines, no comments, etc.

            It's quicker to extend a list than to insert, so add to the end.
            Note: we reverse the original, extend() then reverse again, just
            in case the caller cares.

            nums is a list of linenumbers. Needed to report error lines.
        """

        if not self.que:
            self.que = ['']
            self.qnums = [gbl.lineno]

        q.reverse()
        self.que.extend(q)
        q.reverse()

        nums.reverse()
        self.qnums.extend(nums)
        nums.reverse()

    def read(self):
        """ Return a line.

            This will return either a queued line or a line from the
            file (which was stored/processed earlier).
        """

        while 1:
            if self.que:            # Return a queued line if possible.
                ln = self.que.pop(-1)

                gbl.lineno = self.qnums.pop()

                if not ln:
                    continue

                return ln

            # Return the next line in the file.
            if self.lineptr >= self.lastline:
                return None  # EOF

            ln = self.fdata[self.lineptr].data
            gbl.lineno = self.fdata[self.lineptr].lnum
            self.lineptr += 1

            return ln




# notelen.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>


"""

from . import gbl
from mma.MMA.common import *

noteLenTable = {
    '0': 1,                   # special 0==1 midi tick
    '1': gbl.BperQ * 4,       # whole note
    '2': gbl.BperQ * 2,       # 1/2
    '23': gbl.BperQ * 4 // 3,  # 1/2 triplet
    '4': gbl.BperQ,           # 1/4
    '43': gbl.BperQ * 2 // 3,  # 1/4 triplet
    '8': gbl.BperQ // 2,       # 1/8
    '81': None,               # short 1/8, set by swing skew
    '82': None,               # long 1/8, set by swing skew
    '16': gbl.BperQ // 4,      # 1/16
    '32': gbl.BperQ // 8,      # 1/32
    '64': gbl.BperQ // 16,     # 1/64
    '6': gbl.BperQ // 6,       # 1/16 note triplet
    '3': gbl.BperQ // 3,       # 1/8 note triplet
    '5': gbl.BperQ // 5}       # 1/8 note quintuplet


def getNoteLen(n):
    """ Convert a Note to a midi tick length.

    Notes are 1==Whole, 4==Quarter, etc.
    Notes can be dotted or double dotted.
    Notes can be combined: 1+4 == 5 beats, 4. or 4+8 == dotted 1/4
                           1-4 == 3 beats, 1-0 == 4 beats less a midi tick
    Tuples can be set using (count,base).
    """

    length = 0

    if n.endswith('T') or n.endswith('t'):
        n = n[:-1]
        if not n:
            error("Specifying a note length in MIDI Ticks requires a value")
        n = stoi(n)
        if n == 0:
            n = 1
        if n < 1:
            error("MIDI note length cannot be negative.")
        return n

    n = n.replace('-', '+-')    # change "2-4" to "2+-4" for easier parsing
    n = n.replace('++-', '+-')  # and in case we already used "+-", take out 2nd "+"

    for a in str(n).split('+'):
        if a.endswith('..'):
            dot = 2
            a = a[:-2]
        elif a.endswith('.'):
            dot = 1
            a = a[:-1]
        else:
            dot = 0

        try:
            if a.startswith('-'):
                mult = -1
                a=a[1:]
            else:
                mult = 1
            if ':' in a:
                i = getTupleValue(a)
            else:
                i = noteLenTable[a]
            i *= mult

        except:
            # This also catches things like "4..." 
            error("Unknown note duration %s" % n)

        if dot == 2:
            i += i / 2 + i / 4
        elif dot == 1:
            i += i / 2
        length += i

    return int(length)

def getTupleValue(a):
    """ Grab a tuplet value from note 'xx:yy' We've already verified that
        the string has a ':'
    """

    dd, nn = a.split(':')
    if nn not in ('1', '2', '4', '8', '16', '32', '64'):
        error("Illegal duration value for tuplet. Must be a note "
              "(1,2,4...), not %s." % nn)
    
    nn = noteLenTable[nn]

    # validate the divisor a bit
    dd = stoi(dd)
    if dd<1 or dd>128:
        error("The count value for the tuplet '%s' is out of range." % a)
    return  nn // dd




# alloc.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import mma.MMA.patChord
import mma.MMA.patWalk
import mma.MMA.patBass
import mma.MMA.patPlectrum
import mma.MMA.patDrum
import mma.MMA.patScale
import mma.MMA.patArpeggio
import mma.MMA.patSolo
import mma.MMA.patAria
import mma.MMA.grooves
import mma.MMA.debug

from . import gbl
from mma.MMA.common import *

trkClasses = {
    'BASS'     : mma.MMA.patBass.Bass,
    'CHORD'    : mma.MMA.patChord.Chord,
    'ARPEGGIO' : mma.MMA.patArpeggio.Arpeggio,
    'SCALE'    : mma.MMA.patScale.Scale,
    'DRUM'     : mma.MMA.patDrum.Drum,
    'WALK'     : mma.MMA.patWalk.Walk,
    'MELODY'   : mma.MMA.patSolo.Melody,
    'SOLO'     : mma.MMA.patSolo.Solo,
    'ARIA'     : mma.MMA.patAria.Aria,
    'PLECTRUM' : mma.MMA.patPlectrum.Plectrum

}


def trackAlloc(name, err):
    """ Check existence of track and create if possible.

        If 'err' is set, the function will 'error out' if
        it's not possible to create the track. Otherwise,
        it is content to return without creation taking place.
    """

    # If the track already exists, just return

    if name in gbl.tnames:
        return

    # Get the trackname. Can be just a type, or type-name.

    if '-' in name:
        base, ext = name.split('-', 1)
    else:
        ext = None
        base = name

    """ See if there is a track class 'base'. If there is, then
        'f' points to the initialization function for the class.
        If not, we either error (err==1) or return (err==0).
    """

    try:
        f = trkClasses[base]
    except KeyError:
        if err:
            error("There is no track class '%s' for trackname '%s'" % (base, name))
        else:
            return

    # Now attempt to allocate the track

    gbl.tnames[name] = newtk = f(name)

    # Update current grooves to reflect new track.

    for slot in mma.MMA.grooves.glist.keys():
        newtk.saveGroove(slot)

    if mma.MMA.debug.debug:
        dPrint("Creating new track %s" % name)

    return



# safe_eval.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

# This code pretends to implement a "safe" eval(). It really
# isn't safe! There are many ways an evil coder can exploit
# the safeguards. Unfortunately, python just isn't written to
# have a really bullet proof eval(). Look at your scripts before
# running things blindly!

import re
from math import *
from random import randint
from os import environ
from mma.MMA.common import *

safeCmds = ['ceil', 'fabs', 'floor', 'exp', 'log', 'log10', 'pow',
            'sqrt', 'acos', 'asin', 'atan', 'atan2', 'cos', 'hypot',
            'sin', 'tan', 'degrees', 'radians', 'cosh', 'sinh',
            'int', 'in', '.join', 'str', '.split', 'for', 'randint' ]


def safeEnv(var):
    """ Return the value of an env variable. 
        On my system non-existant env vars register as None and
        vars set to '' return as '' ... so we convert them all to
        ''. MMA doesn't have a NoneType.
    """

    ret = environ.get(var)
    if ret == None:
        ret == ''
    return ret

def safeEval(expr):
    toks = re.split(r'([a-zA-Z_\.]+|.)', expr)

    for t in toks:
        if len(t) > 1 and t not in safeCmds:
            error("Illegal/Unknown operator '%s' in $()." % t)
    try:
        return eval(expr)
    except:
        error("Illegal operation in '%s'." % expr)




# midifuncs.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

Low level entry points, mostly called directly from the parser.
"""

import struct
from . import gbl
import mma.MMA.mdefine
from   mma.MMA.common import *
import mma.MMA.debug

# Storage for midi channel init commands (MidiInit). Commands are added
# from the functions setMidiInt and trackSetMidiInit in this module.
# They are dumped out in parse.py as channels are assigned to tracks.

channelInit = {}
for c in range(1, 17):
    channelInit[c] = []

masterMidiVolume = 0x3fff   # assume device is set to max volume

# non-track functions

def midiMarker(ln):
    """ Parse off midi marker. """

    if len(ln) == 2:
        offset = stof(ln[0])
        msg = ln[1]
    elif len(ln) == 1:
        offset = 0
        msg = ln[0]
    else:
        error("Usage: MidiMark [offset] Label")

    offset = int(gbl.tickOffset + (gbl.BperQ * offset))
    if offset < 0:
        error("MidiMark offset points before start of file")

    gbl.mtrks[0].addMarker(offset, msg)


def setMidiCue(ln):
    """ Insert MIDI cue (text) event into meta track."""

    if not ln:
        error("MidiCue requires text.")

    gbl.mtrks[0].addCuePoint(gbl.tickOffset, ' '.join(ln))


def rawMidi(ln):
    """ Send hex bytes as raw midi stream. """

    mb = []
    for a in ln:    # loop needed to verify, otherwise we'd use list comprehension
        a = stoi(a)
        if a < 0 or a > 0xff:
            error("All values must be in the range 0 to 0xff, not '%s'" % a)
        mb.append(a)
    mb = mma.MMA.midiM.packBytes((mb))   # save value for debug
    gbl.mtrks[0].addToTrack(gbl.tickOffset, mb)

    if mma.MMA.debug.debug:
        dPrint("MIDI: Inserted raw midi in metatrack: %s" %
              ' '.join([str(a) for a in struct.unpack("%sB" % len(mb), mb)]))



def setMidiFileType(ln):
    """ Set some MIDI file generation flags. """

    if not ln:
        error("USE: MidiFile [SMF=0/1] [RUNNING=0/1]")

    for l in ln:
        try:
            mode, val = l.upper().split('=')
        except:
            error("Each arg must contain an '=', not '%s'" % l)

        if mode == 'SMF':
            if val == '0':
                gbl.midiFileType = 0
            elif val == '1':
                gbl.midiFileType = 1
            else:
                error("Use: MIDIFile SMF=0/1")

        elif mode == 'RUNNING':
            if val == '0':
                gbl.runningStatus = 0
            elif val == '1':
                gbl.runningStatus = 1
            else:
                error("Use: MIDIFile RUNNING=0/1")

        else:
            error("Use: MIDIFile [SMF=0/1] [RUNNING=0/1]")

    if mma.MMA.debug.debug:
        if gbl.runningStatus:
            a = 'ON'
        else:
            a = 'OFF'
        dPrint("MIDIFile: SMF=%s RUNNING=%s" % (gbl.midiFileType, a))


def setChPref(ln):
    """ Set MIDI Channel Preference. """

    if not ln:
        error("Use: ChannelPref TRACKNAME=CHANNEL [...]")

    for i in ln:
        if '=' not in i:
            error("Each item in ChannelPref must have an '='")

        n, c = i.split('=')

        c = stoi(c, "Expecting an integer for ChannelPref, not '%s'" % c)

        if c < 1 or c > 16:
            error("Channel for ChannelPref must be 1..16, not %s" % c)

        gbl.midiChPrefs[n.upper()] = c

    if mma.MMA.debug.debug:
        dPrint("ChannelPref: %s" % 
              ' '.join(["%s=%s" % (n, c) for n, c in gbl.midiChPrefs.items()]))


def setMidiCopyright(ln):
    """ Add a copyright message to the file. This is inserted into
        the meta track at offset 0.
    """

    if not ln:
        error("MidiCopyright needs text message.")

    gbl.mtrks[0].addCopyright(0, ' '.join(ln))


def setMidiName(ln):
    """ Set global/meta track name. This will overwrite the song name set in main."""

    if not ln:
        error("Use: TrackName text")

    gbl.mtrks[0].addTrkName(0, ' '.join(ln))


def setMidiText(ln):
    """ Set midi text into meta track."""

    if not ln:
        error("Use: MidiText text")

    gbl.mtrks[0].addText(gbl.tickOffset, ' '.join(ln))


def setMidiCresc(ln):
    doMidiTrackCresc(ln, 1, "MidiCresc")


def setMidiDecresc(ln):
    doMidiTrackCresc(ln, -1, "MidiDeCresc")


def doMidiTrackCresc(ln, dir, func):
    """ Low level MIDI (de)cresc channel volume.

        This is mostly the same as the track function, but since
        we are dealing with a 14bit rather 7bit value it's maybe
        easier to have it separate.
    """

    global masterMidiVolume
    step = 10

    ln, opts = opt2pair(ln, toupper=True)
    for o, v in opts:
        if o == 'STEP':
            step = stoi(v)
            if step < 1:
                error("%s: Step rate must be >0." % func)

        else:
            error("%s: Unknow option '%s'." % (func, o))

    if len(ln) not in (2, 3):
        error("%s: usage -  [<start>] <end> <count>" % (func))

    if len(ln) == 2:
        # No start, insert the current master volume
        ln.insert(0, str(masterMidiVolume))

    v1 = mma.MMA.volume.calcMidiVolume14(ln[0])
    if v1 < 0 or v1 > 0x3fff:
        error("%s: Volume must be 0..16383." % (v1))

    v2 = mma.MMA.volume.calcMidiVolume14(ln[1])
    if v2 < 0 or v2 > 0x3fff:
        error("%s: Volume must be 0..16383." % (v1))

    count = stof(ln[2])
    if count <= 0:
        error("%s: count must be >0" % (func))

    if dir == -1 and v1 < v2:
        warning("%s: dest volume > start" % (func))
    elif dir == 1 and v1 > v2:
        warning("%s: dest volume < start" % (func))

    ldur = int(count * gbl.barLen)
    start = gbl.tickOffset
    end = gbl.tickOffset + ldur

    #if v1 > v2:
    #    v1, v2 = v2, v1
    #volspan = v2 - v1
    if step > ldur:
        step = ldur
    changer = ((v2 - v1 ) // (ldur // step))
    
    v = v1
    t = 0
    for d in range(start, end, step):
        if v < 0 or v > 0x3fff:
            break
        gbl.mtrks[0].addMasterVolume(d, v)
        v += changer
        t += 1

    masterMidiVolume = v2
    gbl.mtrks[0].addMasterVolume(end, v2)

    if mma.MMA.debug.debug:
        dPrint("MIDI(de)Cresc: Added %s changes" % t)


def setMidiVolume(ln):
    """ Set midi master volume."""

    global masterMidiVolume

    if len(ln) != 1:
        error("MidiVolume: Needs exactly one argument.")

    v = mma.MMA.volume.calcMidiVolume14(ln[0])

    if v < 0 or v > 0x3fff:
        error("MidiVolume: Volume setting must be 0...16383.")

    gbl.mtrks[0].addMasterVolume(gbl.tickOffset, v)
    masterMidiVolume = v

    if mma.MMA.debug.debug:
        dPrint("MidiVolume: Master volume set to %s." % v)


def setChannelInit(ln):
    """ Set a command for all midi channels. This is stored and
        added to a track when a channel is assigned. If the channel
        is never assigned, the command never takes effect.
    """

    channels = []

    ln, opts = opt2pair(ln)

    for o, v in opts:
        if o.upper() == 'CHANNELS':
            for c in v.split(','):
                if '-' in c:
                    c1, c2 = c.split('-')
                    c1 = stoi(c1)
                    c2 = stoi(c2)
                    if c2 < c1:
                        c1, c2 = c2, c1
                    if c1 <= 0 or c2 > 16:
                        error("ChannelInit: Valid channels range 1..16, not %s-%s." % (c1, c2))
                    channels.extend(range(c1, c2+1))
                else:
                    c = stoi(c)
                    if c <= 0 or c > 16:
                        error("ChannelInit: Valid channels are 1..16, not %s." % c)
                    channels.append(c)
            channels = list(set(channels))
        else:
            error("ChannelInit: %s is not a valid option." % o)

    if not channels:
        channels = range(1, 17)

    if not len(ln):
        error("ChannelInit: A command is required.")

    ln[0] = ln[0].upper()
    if ln[0] not in mma.MMA.parse.trackFuncs:
        error("ChannelInit: Track function %s does not exist." % ln[0])

    for c in channels:
        channelInit[c].append(ln)

    if mma.MMA.debug.debug:
        dPrint("ChannelInit: '%s' queued to channels %s" % 
            (' '.join(ln), ','.join([str(c) for c in channels])))


def doChannelInit(channel, name):
    """ Add the ChannelInit stuff to the start of this track.
        Pull from the storage and process the string as if it were a line
        from a file. We are calling up functions in the parse module to do this.
        But, parse.py is not imported... can't since it causes a problem with
        trackFuncs table not existing. So, this might cause a future problem!!!
    """

    for l in mma.MMA.midifuncs.channelInit[channel]:
        mma.MMA.parse.trackFuncs[l[0]](name, l[1:])
    mma.MMA.midifuncs.channelInit[channel] = []

################################################
## Track functions


def trackMidiVolume(name, ln):
    """ LowLevel MIDI command. Set Channel Volume. """

    if len(ln) != 1:
        error("%s MidiVolume: Needs exactly 1 arg." % name)

    v = mma.MMA.volume.calcMidiVolume(ln[0])
    if v < 0 or v > 127:
        error("%s MidiVolume: Volumes need to be 0..127, not %s." % (name, v))

    gbl.tnames[name].midiPending.append(("CVOLUME", gbl.tickOffset, v))
    gbl.tnames[name].cVolume = v

    if mma.MMA.debug.debug:
        mma.MMA.debug.trackSet(name, 'MIDIVolume')


def trackMidiCresc(name, ln):
    """ MIDI cresc. """

    doMidiCresc(name, ln, 1, "MidiCresc")


def trackMidiDecresc(name, ln):
    """ MIDI cresc. """

    doMidiCresc(name, ln, -1, "MidiDecresc")


def doMidiCresc(name, ln, dir, func):
    """ Low level MIDI (de)cresc channel volume."""

    tptr = gbl.tnames[name]

    if len(ln) not in (2, 3):
        error("%s %s: usage -  [<start>] <end> <count>" % (name, func))

    if len(ln) == 2:
        # If passing only one param we're doing a change from the
        # current volume. Unfort, that setting might be buffered
        # in a channelinit list or in the pending queue. So, we
        # force a channel, force init stuff and check the pending
        # buffer (from the end).
        
        if not tptr.channel:
            tptr.setChannel()
        doChannelInit(tptr.channel, name)
        currentVol = tptr.cVolume
        for c, off, v in reversed(tptr.midiPending):
            if c == 'CVOLUME':
                currentVol = v
                break

        ln.insert(0, str(currentVol))

    v1 = mma.MMA.volume.calcMidiVolume(ln[0])
    if v1 < 0 or v1 > 127:
        error("%s %s: Volume must be 0..127." % (name, v1))
    v2 = mma.MMA.volume.calcMidiVolume(ln[1])
    if v2 < 0 or v2 > 127:
        error("%s %s: Volume must be 0..127." % (name, v1))

    count = stof(ln[2])
    if count <= 0:
        error("%s %s: count must be >0" % (name, func))

    if dir == -1 and v1 < v2:
        warning("%s %s: dest volume > start" % (name, func))
    elif dir == 1 and v1 > v2:
        warning("%s %s: dest volume < start" % (name, func))

    t = abs(v2-v1)
    step = (count * gbl.barLen) // t  # step rate
    p = gbl.tickOffset

    if v2 < v1:
        dir = -1
    else:
        dir = 1

    for v in range(v1, v2+dir, dir):
        gbl.tnames[name].midiPending.append(("CVOLUME", int(p), v))
        p += step

    if mma.MMA.debug.debug:
        dPrint("%s MidiVolume: Added %s changes" % (name, t))

def trackGlis(name, ln):
    """ Enable/disable portamento. """

    if len(ln) != 1:
        error("%s MidiGlis: NN, off=0, 1..127==on" % name)

    v = stoi(ln[0], "Expecting integer for Portamento")

    if v < 0 or v > 127:
        error("MidiGlis %s: Value for Glis/Portamento must be 0..127." % name)

    gbl.tnames[name].midiPending.append(("GLIS", gbl.tickOffset, v))

    if mma.MMA.debug.debug:
        dPrint("Set %s MIDIGlis to %s" % (name, v))


def trackWheel(name, ln):
    """ Set a pitch bend (wheel) set for a given track. """

    tdata = gbl.tnames[name]

    startOffset = None
    endOffset = 0
    setOnly = None
    duration = None
    startValue = None
    endValue = None
    step = 10
    reset = True
    cycle = False
    rate = None

    ln, opts = opt2pair(ln, toupper=True)

    if ln:
        if len(ln) == 1 and ln[0] == 'RESET':
            opts.append(('SET', 'CENTER'))
        else:
            error("%s MidiWheel: Unrecognized command(s) '%s'." % (name, ' '.join(ln)))

    for o, v in opts:
        if o == 'OFFSET':
            if v.endswith('M'):
                mul = gbl.QperBar
                v = v[:-1]
            else:
                mul = 1
            startOffset = gbl.tickOffset + (int(stof(v) * mul * gbl.BperQ))
            if startOffset < 0:
                error("%s MidiWheel Offset: start is before song start." % name)

        elif o == 'DURATION' or o == 'COUNT':
            if v.endswith('M'):
                mul = gbl.QperBar
                v = v[:-1]
            else:
                mul = 1
            duration = int(stof(v) * mul * gbl.BperQ)
            if duration <= 0:
                error("%s MidiWheel: Duration must be > 0, not '%s'." %
                      (name, duration))

        elif o == 'SET':
            if v == 'CENTER':
                setOnly = 0x2000
            else:
                setOnly = stoi(v)

        elif o == 'START':
            if v == 'CENTER':
                startValue = 0x2000
            else:
                startValue = stoi(v)

        elif o == 'END':
            if v == 'CENTER':
                endValue = 0x2000
            else:
                endValue = stoi(v)

        elif o == 'STEP':
            step = stoi(v)
            if step < 1:
                error("%s MidiWheel Step: must be > 0." % name)

        elif o == 'RESET':
            if v in ('NO', '0', 'OFF'):
                reset = 0
            elif v in ('YES', '1', 'ON'):
                reset = 1
            else:
                error("%s MidiWheel Reset: Use ON or OFF, not '%s'." % (name, v))

        elif o == 'RATE':
            rate = mma.MMA.notelen.getNoteLen(v)

        elif o == 'CYCLE':
            if v in ('NO', '0', 'OFF'):
                cycle = False
            elif v in ('YES', '1', 'ON'):
                cycle = True
            else:
                error("%s MidiWheel Cycle: Use ON or OFF, not '%s'." % (name, v))
        else:
            error("%s MidiWheel: Unrecognized command '%s'." % (name, o))

    if setOnly is not None:
        if len(opts) > 2 or len(opts) > 1 and startOffset is None:
            warning("MidiWheel %s Set: some options may be ignored." % name)
        if startOffset is None:
            startOffset = gbl.tickOffset
        tdata.midiPending.append(("WHEEL", startOffset, setOnly))

        if mma.MMA.debug.debug:
            dPrint("MidiWheel %s: detuned to %s at %s ticks." %
                (name, setOnly, startOffset))
        return

    if startOffset is None:
        startOffset = gbl.tickOffset

    if duration is None:
        error("%s MidiWheel: Duration must be set." % name)

    endOffset = startOffset + duration

    if startValue is None or endValue is None:
        error("%s MidiWheel: No start/end value set." % name)

    if startValue < 0 or startValue > 16383:
        error("%s MidiWheel: Start must be 0 ... 16383, not '%s'." % (name, startValue))

    if endValue < 0 or endValue > 16383:
        error("%s MidiWheel: End must be 0 ... 16383, not '%s'." % (name, endValue))

    if rate is None:
        repeat = 1
        ldur = duration
    else:
        ldur = rate              # duration of each loop
        repeat = duration // ldur  # times to loop

    if startValue < endValue:
        pitchspan = endValue - startValue
    else:
        pitchspan = startValue - endValue

    if step > ldur:
        step = ldur
    changer = pitchspan // (ldur/step)

    if startValue > endValue:
        changer *= -1

    value = startValue
    off = startOffset

    for r in range(repeat):
        for d in range(0, ldur, step):
            if value < 0:
                value = 0
            if value > 0x3fff:
                value = 0x3fff

            tdata.midiPending.append(("WHEEL", int(off+d), int(value)))
            value += changer
        off += ldur
        if cycle:
            changer *= -1

        if not cycle:
            value = startValue

    if reset:
        tdata.midiPending.append(("WHEEL", off+5, 0x2000))  # reset at end

    if mma.MMA.debug.debug:
        if not reset:
            rset = "No RESET." 
        else:
            rset = ''
        dPrint("MidiWheel %s: detuned %s to %s from %s to %s. %s" % 
            (name, startValue, endValue, startOffset, endOffset, rset))


# Lookup table constants for MIDI PAN values.
panNames = {'LEFT': 0,       'LEFT100': 0,   'LEFT90': 6,
            'LEFT80': 13,    'LEFT70': 19,   'LEFT60': 25,
            'LEFT50': 31,    'LEFT40': 39,   'LEFT30': 44,
            'LEFT20': 50,    'LEFT10': 57,   'CENTER': 64,
            'RIGHT10': 70,   'RIGHT20': 77,  'RIGHT30': 83,
            'RIGHT40': 88,   'RIGHT50': 96,  'RIGHT60': 102,
            'RIGHT70': 108,  'RIGHT80': 114, 'RIGHT90': 121,
            'RIGHT100': 127, 'RIGHT': 127 }

def trackPan(name, ln):
    """ Set the Midi Pan value for a track."""

    def getv(v):
        try:
            return panNames[v.upper()]
        except:
            return stoi(v, "Expecting integer value 0..127 or mnemonic (Left*, Center, Right*).")


    tdata = gbl.tnames[name]

    if len(ln) not in (1, 3):
        error("MidiPan %s: needs 1 arg [Value] OR 3 [Initvalue DestValue Beats/Measures]." % name)


    if len(ln) == 3:
        if ln[2].upper().endswith('M'):
            beats = int(stof(ln[2][:-1]) * gbl.QperBar)
        else:
            beats = stof(ln[2])
        if beats < 1:
            error("MidiPan %s: Beat/Measure value must be positive count, "
                  "not '%s'." % (name,beats))
        initPan = getv(ln[0])
        newPan = getv(ln[1])
    else:
        beats = 0
        initPan = 0
        newPan = getv(ln[0])

    ticks = beats * gbl.BperQ   # convert beats to midi ticks

    if newPan < 0 or newPan > 127:
        error("MidiPAN: final value must be 0..127")

    if newPan < initPan:
        span = initPan-newPan
        changer = -1
    else:
        span = newPan-initPan
        changer = 1

    if span > 0:
        step = ticks/span
    else:
        beats = 0

    if beats:
        v = initPan
        off = gbl.tickOffset
        for a in range(span+1):
            tdata.midiPending.append(("PAN", int(off), v))
            off += step
            v += changer

    else:
        tdata.midiPending.append(("PAN", gbl.tickOffset, newPan))

    if mma.MMA.debug.debug:
        if beats:
            dPrint("Set %s MIDIPan from %s to %s over %s beats." % 
                (name, initPan, newPan, beats))
        else:
            dPrint("Set %s MIDIPan to %s" % (name, newPan))


def trackMidiText(name, ln):
    """ Insert midi text event. """

    if not ln:
        error("Use: %s Text" % name)

    ln = ' '.join(ln)
    gbl.tnames[name].midiPending.append(("MIDITEXT", gbl.tickOffset, ln))

    if mma.MMA.debug.debug:
        dPrint("Set %s MIDIText '%s'." % (name, ln))


def trackMidiCue(name, ln):
    """ Insert MIDI cue (text) event."""

    if not ln:
        error("MidiCue %s: Needs arg(s)." % name)

    ln = ' '.join(ln)
    gbl.tnames[name].midiPending.append(("MIDICUE", gbl.tickOffset, ln))

    if mma.MMA.debug.debug:
        dPrint("Set %s MIDICue '%s'." % (name, ln))


def trackMidiExt(ln):
    """ Helper for trackMidiSeq() and trackMidiVoice()."""

    ids = 1
    while 1:
        sp = ln.find("{")

        if sp < 0:
            break

        ln, s = pextract(ln, "{", "}", 1)
        if not s:
            error("Did not find matching '}' for '{'")

        pn = "_%s" % ids
        ids += 1

        mma.MMA.mdefine.mdef.create(pn, s[0])
        ln = ln[:sp] + ' ' + pn + ' ' + ln[sp:]

    return ln.split()


def trackMidiSeq(name, ln):
    """ Set reoccurring MIDI command for track. """

    if not ln:
        error("Use %s MidiSeq Controller Data" % name)

    if ln[0][0] != '{':      # add {} wrapper if missing
        ln.insert(0, '{')
        ln.extend('}')

    if len(ln) == 1 and ln[0] == '-':
        gbl.tnames[name].setMidiSeq('-')
    else:
        gbl.tnames[name].setMidiSeq(trackMidiExt(' '.join(ln)))


def trackMidiVoice(name, ln):
    """ Set single shot MIDI command for track. """

    if not ln:
        error("Use %s MidiVoice Controller Data" % name)

    if ln[0][0] != '{':      # add {} wrapper if missing
        ln.insert(0, '{')
        ln.extend('}')

    if len(ln) == 1 and ln[0] == '-':
        gbl.tnames[name].setMidiVoice('-')
    else:
        gbl.tnames[name].setMidiVoice(trackMidiExt(' '.join(ln)))


def trackMidiClear(name, ln):
    """ Set MIDI command to send at end of groove. """

    if not ln:
        error("Use %s MIDIClear Controller Data" % name)

    if len(ln) == 1 and ln[0] == '-':
        gbl.tnames[name].setMidiClear('-')
    else:
        ln = ' '.join(ln)
        if '{' in ln or '}' in ln:
            error("{}s are not permitted in %s MIDIClear command" % name)
        gbl.tnames[name].setMidiClear(trackMidiExt('{' + ln + '}'))


def trackMidiName(name, ln):
    """ Set channel track name."""

    if len(ln) != 1:
        error("MidiTName %s: Use exactly one arg." % name)

    gbl.tnames[name].midiPending.append(('TNAME', 0, ln[0]))

    if mma.MMA.debug.debug:
        dPrint("Set %s MIDI Track Name to %s" % (name, ln[0]))



# common.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>


These are a collection of miscellaneous routines used in various
parts of mma.MMA. It is safe to load the whole works with:

    from mma.MMA.common import *

without side effects (yeah, right).

"""
from random import randrange
import sys
import time
from . import gbl
from textwrap import wrap
import mma.MMA.debug

# A buffer for various debug/warning/error messages
# this is only used if MMA_LOGFILE has been set. The
# buffer is dumped at exit.
outBuffer = []

# having the term width is nice for pretty print error/warning
from mma.MMA.termsize import getTerminalSize
termwidth = getTerminalSize()[0]-1

def bufferPrint(a):
    if gbl.logFile:
        outBuffer.append(a)
    else:
        print(a)

def cleanPrintBuffer():
    """ Write an existing stored buffer (debug/error/warning). """
    
    if outBuffer and gbl.logFile:
        outBuffer.insert(0, "\n**** Run log: '%s' %s" % ( gbl.infile, time.asctime()))
        
        opath = open(gbl.logFile, 'a')    # Open output for append
        try:
            import fcntl
            fcntl.flock(opath, fcntl.LOCK_EX)   # make sure we print in one batch
        except:
            pass
        opath.write('\n'.join(outBuffer))   # dump entire buffer
        opath.close()                       # this will release lock as well

def prettyPrint(msg):
    """ Simple formatter for error/warning messages."""

    if isinstance(msg, list):
        msg = ' '.join(msg)
    try:
        for a in wrap(msg, termwidth, initial_indent='', subsequent_indent='    '):
            bufferPrint(a)
    except:
        bufferPrint(msg)

def error(msg):
    """ Print an error message and exit.

        If the global line number is >=0 then print the line number
        as well.
    """

    if gbl.lineno >= 0:
        linno = "<Line %d>" % gbl.lineno
    else:
        linno = ''

    if gbl.inpath:
        file = "<File:%s>" % gbl.inpath.fname
    else:
        file = ''

    prettyPrint("Error: %s %s %s" % (linno, file, msg))

    # Parse though the error message and check for illegal characters.
    # Report (first only) if any found.

    for a in msg:
        a = ord(a)
        if a < 0x20 or a >= 0x80:
            bufferPrint("Corrupt input file? Illegal character 'x%02x' found." % a)
            break

    if gbl.ignoreBadChords:  # set with -xCHORD and -xPERMISSIVE options
        return
    
    sys.exit(1)

def warning(msg):
    """ Print warning message and return. """

    if not mma.MMA.debug.noWarn:
        if gbl.lineno >= 0:
            linno = "<Line %d>" % gbl.lineno
        else:
            linno = ''

        if gbl.inpath:
            file = "<File:%s>" % gbl.inpath.fname
        else:
            file = ''

        prettyPrint("Warning: %s %s %s" % (linno, file, msg))


def dPrint(msg):
    """ Print/buffer a debugging message. Keep separate since we might
        want to install line numbering, etc. later???
    """

    prettyPrint(msg)
    
def getOffset(ticks, ranLow=None, ranHigh=None):
    """ Calculate a midi offset into a song.

        ticks == offset into the current bar.
        ran   == random adjustment from RTIME

        When calculating the random factor the test ensures
        that a note never starts before the start of the bar.
        This is important ... voice changes, etc. will be
        buggered if we put the voice change after the first
        note-on event.

        CAUTION: If you pass either ranLow or ranHigh make sure
                 both are set. If one has a value and the other
                 is None a crash is ensured. The existing callers
                 are okay (setRtime creates both values).
    """

    p = gbl.tickOffset + int(ticks)     # int() cast is important!

    if ranLow or ranHigh:
        r = randrange(ranLow, ranHigh+1)
        if ticks == 0 and r < 0:
            r = 0
        p += r

    return p


def stoi(s, errmsg=None):
    """ string to integer. """

    try:
        return int(s, 0)
    except ValueError:
        if errmsg:
            error(errmsg)
        else:
            error("Expecting integer value, not %s" % s)


def stof(s, errmsg=None):
    """ String to floating point. """

    try:
        return float(s)
    except:
        try:
            return int(s, 0)
        except ValueError:
            if errmsg:
                error(errmsg)
            else:
                error("Expecting a  value, not %s" % s)


def pextract(s, open, close, onlyone=None, insert=''):
    """ Extract a parenthesized set of substrings.

        s        - original string
        open    - substring start tag \ can be multiple character
        close   - substring end tag   / strings (ie. "<<" or "-->")
        onlyone - optional, if set only the first set is extracted
        insert  - optional, insert string where extraction

        returns ( original sans subs, [subs, ...] )

        eg: pextract( "x{123}{666}y", '{',    '}' )
            Returns:  ( 'xy', [ '123', '666' ] )

    """

    subs = []
    magic = chr(1)
    while 1:
        lstart = s.find(open)
        lend   = s.find(close)

        if lstart > -1 and lstart < lend:
            subs.append(s[lstart + len(open):lend].strip())
            s = s[:lstart] + magic + s[lend+len(close):]
            if onlyone:
                break
        else:
            break

    s = s.replace(magic, insert)
    
    return s.strip(), subs


def seqBump(l):
    """ Expand/contract an existing sequence list to the current seqSize."""

    while len(l) < gbl.seqSize:
        l.extend(l)
    return l[:gbl.seqSize]


def lnExpand(ln, msg):
    """ Validate and expand a list passed to a set command. """
    
    if len(ln) > gbl.seqSize:
        if not gbl.inAllGrooves:
            warning("%s list truncated to %s patterns" % (msg, gbl.seqSize) )
        ln = ln[:gbl.seqSize]

    last = None

    for i, n in enumerate(ln):
        if n == '/':
            if last is None:
                error("%s cannot use a '/' as the first item in list." % msg)
            else:
                ln[i] = last
        else:
            last = n

    return ln


def opt2pair(ln, toupper=False, notoptstop=False):
    """ Parse a list of options. Separate out "=" option pairs.

        Returns:
           newln - original list stripped of opts
           opts  - list of options. Each option is a tuple(opt, value)

       Note: default is to leave case alone. Setting toupper converts 
                everything to upper.
             default is to parse entire line. Setting notoptstop stops parse at first
                word which is not a xx=yy pair.
    """

    opts = []
    newln = []

    # Permit the user to use stuff like "Beats = 1 , 3, 4" or "opt1 = 44 opt2 = 99"
    # So, join the line with space delminters, then strip out spaces before/after
    # all '=' and ','. This may bugger up print statements, but they aren't parsed
    # here ... so this should be fine. This might be a cause of future errors!

    # This is very useful when the user wants to use a macro expansion:
    #     Beats = $Macro1 , $Macro2
    # since macros need to have spaces around them.

    ln = ' '.join(ln)
    ln = ln.replace('  ', ' ')  # strip double spaces
    ln = ln.replace('= ', '=')  # spaces after/before =
    ln = ln.replace(' =', '=')
    ln = ln.replace(', ', ',')  # spaces after/before ,
    ln = ln.replace(' ,', ',')
    ln = ln.split()

    for v, a in enumerate(ln):
        if toupper:
            a = a.upper()
        try:
            o, v = a.split('=', 1)
            opts.append((o, v))
        except ValueError:   # this means no '=', split() failed
            newln.append(a)
            if notoptstop:
                newln.extend(ln[v+1:])
                break

    return newln, opts



# patWalk.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""


import random


import mma.MMA.harmony
import mma.MMA.notelen

from . import gbl
from mma.MMA.common import *
from mma.MMA.pat import PC, Pgroup


class Walk(PC):
    """ Pattern class for a walking bass track. """

    vtype = 'WALK'
    walkChoice = 0

    def getPgroup(self, ev):
        """ Get group for walking bass pattern.

            Fields - start, length, volume
        """

        if len(ev) != 3:
            error("There must be at exactly 3 items in each group in "
                  "a Walking Bass definition, not <%s>" % ' '.join(ev))

        a = Pgroup()

        a.offset = self.setBarOffset(ev[0])
        a.duration = mma.MMA.notelen.getNoteLen(ev[1])
        a.vol = stoi(ev[2], "Type error in Walking Bass definition")

        return a

    def restart(self):
        self.ssvoice = -1
        self.walkChoice = 0

    def trackBar(self, pattern, ctable):
        """ Do a waling     bass bar.

        Called from self.bar()

        """

        sc = self.seq
        dir = self.direction[sc]

        for p in pattern:

            tb = self.getChordInPos(p.offset, ctable)

            if tb.walkZ:
                continue

            # Create a note list from the current scale. We do
            # this for each beat, but it's pretty fast. The note
            # list is simply notes 0..5 of the scale PLUS notes
            # 1..4 reversed. So, a Cmajor chord would result in
            # the note list (0,2,4,5,7,9,7,5,4,2). We never use
            # scale notes past this point. So in a C chord we
            # are using notes C, D, E, F, G and A. This is one
            # reason we ignore the RANGE setting ... there would
            # be a big gap between the 'A' and the next 'C'.
            #
            # Note that we deliberately skip the 7th. Too often
            # the chord is a Major but the melody note will be
            # the dom. 7th and the M7 will sound off. So, just
            # err on the side of caution.
            #
            # If DIR is UP or DOWN we don't append the 2nd half
            # of the scale.
            #
            # If DIR is DOWN we reverse the order as well.

            wNotes = list(tb.chord.scaleList[0:6])

            if dir not in ('UP', 'DOWN'):
                b = list(tb.chord.scaleList[1:5])
                b.reverse()
                wNotes += b

            if dir == 'DOWN':
                wNotes.reverse()

            # Ensure that the offset is in range.

            if self.walkChoice >= len(wNotes) or self.walkChoice < 0:
                self.walkChoice = 0

            """ Even with a walking bass it's nice to have the chord root on
                beat 1 ... not all the time, but most. This bit of code ensures
                that more that 50% of beat ones will have the root.
            """

            if p.offset == 0 and random.choice((0, 1)):
                self.walkChoice = 0

            note = wNotes[self.walkChoice]

            """ Adjust offset for NEXT TIME. If the direction is
                up/down we just increment the pointer. If we have
                direction set to RANDOM then we select either -1,
                0 or 1 with equal change for moving up, down or
                not-at-all. With BOTH we have a preference to move up.
            """

            if dir in ('UP', 'DOWN'):
                self.walkChoice += 1
            elif dir == 'RANDOM':
                self.walkChoice += random.choice((0, 1, -1))
            else:    # BOTH
                self.walkChoice += random.choice((-1, 0, 0, 2, 2, 1, 1, 1, 1, 1, 1, 1))

            if not self.harmonyOnly[sc]:
                notelist = [(note, p.vol)]
            else:
                notelist = []

            if self.harmony[sc]:
                ch = self.getChordInPos(p.offset, ctable).chord.noteList
                h = mma.MMA.harmony.harmonize(self.harmony[sc], note, ch)
                vol = p.vol * self.harmonyVolume[sc]
                harmlist = list(zip(h, [vol] * len(h)))
            else:
                harmlist = []

            offset = p.offset
            if self.ornaments['type']:
                offset = mma.MMA.ornament.doOrnament(self, notelist,
                                self.getChordInPos(offset, ctable).chord.scaleList, p)
                notelist = []

            self.sendChord(notelist + harmlist, p.duration, offset)



# patPlectrum.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>
Louis James Barman

"""


import mma.MMA.notelen

from . import gbl
from mma.MMA.common import *
from mma.MMA.pat import PC, Pgroup
import mma.MMA.debug 
import copy

class PlecStruct:
    pass

class Fnoise():   # fretnoise options
    def __init__(self):
        self.track = None
        self.strings = []
        self.duration = 192   # 1/4 note default duration in ticks
        self.octave = 0  # octave adjustment
        self.max = 1  # max number of noises per beat
        self.beats = []  # beat filter, []=all
        self.bars = []   # bar filter, []=all

class Plectrum(PC):
    """ Pattern class for a Raw MIDI track. """

    vtype = 'PLECTRUM'

    def __init__(self, nm):
        PC.__init__(self, nm)

        # We have vibrating strings (a string in python refers to text not a guitar string)
        self._vibrating = []
        self._lastOff = []
        self._tuning = []
        self.strumCenter = 0  # default to use 'start'
        self._capoFretNo = 0  # The number that the capo is on (0 for open strings)
        self.setPlectrumTuning(['e-', 'a-', 'd', 'g', 'b', 'e+'])
        self.fretNoise = None
        self.lastChord = None
        self._chordShapes = {}

    def saveGroove(self, gname):
        """ Save special/local variables for groove. """

        PC.saveGroove(self, gname)  # create storage. Do this 1st.
        self.grooves[gname]['CAPO'] = self._capoFretNo
        self.grooves[gname]['TUNING'] = self._tuning[:]
        self.grooves[gname]['PSTRUM'] = self.strumCenter

    def restoreGroove(self, gname):
        """ Restore special/local/variables for groove. """

        self._capoFretNo = self.grooves[gname]['CAPO']
        self._tuning = self.grooves[gname]['TUNING'][:]
        self.strumCenter = self.grooves[gname]['PSTRUM']
        PC.restoreGroove(self, gname)

    def clearSequence(self):
        """ Set some initial values. Called from init and clear seq. """

        PC.clearSequence(self)
        self._capoFretNo = 0
        if self.channel != 0:  # not sure if this is nesc. But safer!
            self.grooveFinish(0)

    def doMidiClear(self):
        """ Reset MIDI settings, special hook for stopping strings. """

        if self.channel != 0:
            self.grooveFinish(0)
        PC.doMidiClear(self)

    def getFretNoiseOptions(self):
        """ Return options. Used in setPlectrumFretNoise() and macros. """

        o = self.fretNoise

        if o.bars:
            bars = ','.join([str(i + 1) for i in o.bars])
        else:
            bars = 'ALL'

        if not o.beats:
            beats = 'All'
        else:
            beats = ','.join([str(1 + (i / float(gbl.BperQ))) for i in o.beats])

        return "%s FretNoise Track=%s Duration=%sT Octave=%s Strings=%s Max=%s Beats=%s Bars=%s" \
            % (self.name, o.track, o.duration, 
             o.octave / 12, 
               ','.join([str(abs(x-len(self._tuning))) for x in o.strings]), 
             o.max,
             beats, bars )


    def setPlectrumFretNoise(self, l):
        """ Set up the fretnoise stuff. """
        
        if not l or len(l)==1 and ( l[0].upper()=='OFF' or l[0].upper()=='NONE'):
            self.fretNoise = None
            if mma.MMA.debug.debug:
                dPrint("%s FretNoise Off" % self.name)
            return

        self.fretNoise = Fnoise()  # new fret noise object
        scount = len(self._tuning)

        notopt, optpair = opt2pair(l, toupper=True)

        if notopt:
            error("%s FretNoise: All options are Key=Value, not '%s'." % \
                  (self.name, ' '.join(notopt)))

        o= self.fretNoise
        for cmd, opt in optpair:

            if cmd == 'BEATS':
               o.beats = []
               if opt != 'ALL':
                   for t in opt.split(','):
                       z = self.name  # this is for error msg in setBarOffset()
                       self.name = "%s FretNoise Beats" % z
                       o.beats.append(self.setBarOffset(t))
                       self.name = z

            elif cmd == 'BARS':
                o.bars = []
                if opt != 'ALL':
                    for t in opt.split(','):
                        i = stoi(t)
                        if i < 1 or i > gbl.seqSize:
                            warning("%s FretNoise Bars: setting of %s may be "
                                "ignored since SeqSize is %s." %
                                (self.name, i, gbl.seqSize))
                        o.bars.append(i - 1)

            elif cmd == 'DURATION':
                o.duration = mma.MMA.notelen.getNoteLen(opt)

            elif cmd == 'MAX':
                i = stoi(opt)
                if i < 1 or i > scount:
                    error("%s FretNoise Max must be 1..%s, not '%s'." % \
                          (self.name, scount, opt))
                o.max = i

            elif cmd == 'STRINGS':
                o.strings = []
                for i in opt.split(','):
                    #i = stoi(i) - 1
                    i = abs(stoi(i)-scount)  # convert 1-6 to 6-1
                    if i < 0 or i > scount:
                        error("%s FretNoise STRING must be 1-%s, not %s." % \
                              (self.name, scount, i))
                    if i not in o.strings:
                        o.strings.append(i)

            elif cmd == 'TRACK':
                mma.MMA.alloc.trackAlloc(opt, 1)
                if gbl.tnames[opt].vtype != 'BASS':
                    error("%s FretNoise Track must be a BASS track, not %s." % \
                          (self.name, gbl.tnames[opt].vtype))
                o.track = opt
                if not gbl.tnames[opt].channel:
                    gbl.tnames[opt].setForceOut()

            elif cmd == 'OCTAVE':
                i = stoi(opt)
                if i < -8 or i > 8:
                    error("%s FretNoise Octave: adjustment must be -8 to 8, not '%s'." 
                          % (self.name, i))
                o.octave = i * 12

            else:
                error("%s FretNoise: Unknown option '%s'." % (self.name, cmd))


        # no strings specified, set to the top 3
        if not o.strings:
            o.strings = range(0, 3)
            # Adjust to fit tuning
            o.strings=o.strings[0:scount]
                
        if not o.track:
            error("%s Fretnoise: No track was set. Use 'Track=some-bass-track' option." % self.name)

        if mma.MMA.debug.debug:
            dPrint(self.getFretNoiseOptions())
        

    def setStrum(self, l):
        """ Called frm parser CENTER option. Sets strum to bar center, start or end.

            NOTE: This overrides the class setStrum() function.
        """

        if len(l) != 1:
            error("Strum: %s permits exactly one strum setting" % self.name)

        l = l[0].lower()

        if l == 'start':
            self.strumCenter = 0
        elif l == 'center':
            self.strumCenter = 1
        elif l == 'end':
            self.strumCenter = 2
        else:
            error("Strum: %s options are 'Start', 'Center' or 'End'. '%s' is illegal" %
                  (self.name, l.title()))

    def setPlectrumTuning(self, stringPitchNames):
        """ This is called from the parser TUNING option to set the
            instrument tuning.

            For standard guitar tuning use setTuning("e- a- d g b e+")

        """
        
        if self.channel != 0:   # if tuning changes while strings are still sounding
            self.grooveFinish(0)

        self._tuning = []
        self._vibration = []
        self.fretNoise = None  # turn off fret noise (could be string count diff)

        for pitchName in stringPitchNames:
            midiPitch = noteNameToMidiPitch(pitchName)
            if midiPitch is None:
                error("%s TUNING: Illegal/unknown string name '%s'." % (self.name, pitchName))

            self._tuning.append(midiPitch)
            vibration = PlecStruct()
            vibration.note = None       # None means the string not vibrating
            vibration.offset = None 
            self._vibrating.append(vibration)

        self._lastOff = [None] * len(self._vibrating)

    def setPlectrumCapo(self, capoFretNo):
        """ Set a capo value. Called from main parser. """

        self._capoFretNo = stoi(capoFretNo, "%s Capo: expecting integer, not '%s'."
                                % (self.name, capoFretNo))

    def setPlectrumShape(self,shape):
        """Defines a chord shape.

            PLECTRUM SHAPE [ chord [ f1 f2 ... fn ] ]

            Defines a shape (fingering) for a specific chord.
            The number of values supplied must match the number
            of strings defined in the current tuning.

            If a shape is defined for a particular chord, this shape
            will be used for the notes to sound. Capo and barre
            settings will be ignored for this chord.

            Allowable values are integers between 0 and 24, and a
            minus sign (or 'x' or 'n') for a muted string.

            Without arguments, removes all current shape definitions.
            With only the chord name, removes the shape definition
            for this chord.

        """

        if len(shape) == 0:
            # Forget all definitions.
            self._chordShapes = {}
            return
        
        chord = shape.pop(0)
        if len(shape) == 0:
            # Forget definition for this chord.
            if chord in self._chordShapes:
                self._chordShapes.pop(chord)
            else:
                warning("%s: Chord %s was not shaped" % (self.name, chord))
            return

        if len(shape) != len(self._tuning):
            error("%s: Shape requires values for %d strings (%d provided)" %
                  (self.name, len(self._tuning), len(shape)))

        self._chordShapes[chord] = []
        for i, n in enumerate(shape):
            if n == '-' or n.lower() == 'x' or n.lower == 'n':
                self._chordShapes[chord].append(None)
            else:
                try:
                    pos = int(n)
                except:
                    pos = None
                if pos == None or pos < -127 or pos > 127:
                    error("%s: Shape requires values between -127 and 127, or '-' for a muted string" % (self.name))
                self._chordShapes[chord].append(pos)
                
    def decodePlectrumPatterns(self, a, patterns):
        """ Decode plectrum patterns for a guitar here are examples """

        a.pluckVol = []
        if patterns[0].find(':') != -1 or len(patterns) == 1:
            # set all strings to note plucked
            for stringNo in range(len(self._tuning)):
                a.pluckVol.append(-1)

            if len(patterns) == 1 and patterns[0].find(':') == -1:
                # put in the missing : if there is just one pattern
                patterns[0] = ":" + patterns[0]

            for patString in patterns:
                if patString.find(':') == -1:
                    error("%s: Not all string definitions have a : in them '%s'" %
                          (self.name, ' '.join(patterns)))

                start = 0
                end = len(self._tuning) - 1

                pat = patString.split(':')

                if pat[0]:
                    emsg = "%s: Note String Number in Plectrum definition not int" % self.name
                    if pat[0].find('-') != -1:
                        startString, endString = pat[0].split('-')
                        if startString:
                            start = stoi(startString, emsg) - 1
                        if endString:
                            end = stoi(endString, emsg) - 1
                    else:
                        start = stoi(pat[0], emsg) - 1
                        end = start

                if start > end:
                    start, end = end, start

                for n in range(start, end + 1):
                    # we number strings from the other end
                    stringNo = len(self._tuning) - n - 1
                    if stringNo < 0 or stringNo >= len(self._tuning):
                        error("%s: string number %d does not exists" % (self.name, stringNo + 1))
                        return
                    if a.pluckVol[stringNo] != -1:
                        error("%s: Duplicate string %d definition" % (self.name, stringNo + 1))
                        return

                    a.pluckVol[stringNo] = stoi(pat[1], "%s: Note volume not int" % self.name)

        else:
            # this is without a the string specifier so all strings must be listed.
            if len(patterns) != len(self._tuning):
                error("%s: There must be %s strings listed definition, "
                      "not '%s'" % (self.name, len(self._tuning), ' '.join(patterns)))
                return
            for stringNo in range(len(self._tuning)):
                val = patterns[stringNo]
                if val == '-':
                    val = '-1'   # -1 means ignore this string
                a.pluckVol.append(stoi(val, "%s: Note volume not int" % self.name))

    def getPgroup(self, ev):
        """ Get group for rawmid pattern.

            Fields - start, length, note, volume

        """

        if len(ev) < 3:  # we need offset, strum and at least one pattern
            error("%s: There must be n groups of 3 or more in a pattern definition, "
                  "not '%s'" % (self.name, ' '.join(ev)))

        a = Pgroup()
        a.vol = 0        # as is this
        a.offset = self.setBarOffset(ev[0])
        a.strum = stoi(ev[1], "%s: Expecting int value for strum" % self.name)

        self.decodePlectrumPatterns(a, ev[2:])

        """ For the doc generators ... we need a setting for duration, even
            though this track doesn't really have one. Using the value of 0
            pretty much matches the results for a drum track. The author of
            this file used "pluckVol" and the docs are expecting a "vol" variable
            so we just duplicate it ... easier than changing all the uses here.
        """

        a.duration = 0   # this is a dummy value to keep docs happy
        a.vol = a.pluckVol

        return a

    def restart(self):
        self.ssvoice = -1

    def doFretNoise(self, stringNo, note, offset):
        """ Add in a fret noise note. """

        if self._lastOff[stringNo] is None or \
           self._lastOff[stringNo].note == note:
            return

        trk = gbl.tnames[self.fretNoise.track]

        # Let user override fretnoise for track by setting a RIFF,
        # or TRIGGER is the bass track.
        if trk.disable or \
           trk.riff or \
           trk.trigger.mode:
            return

        o = self.fretNoise

        # skip if incorrect beat/bar setting
        if (o.beats and offset not in o.beats) \
            or (o.bars and gbl.seqCount not in o.bars):
            return

        # use the plectrum track's seq (the BASS doesn't set it)
        trk.seq = self.seq

        # Use same volume as the string on
        volume = self._lastOff[stringNo].volume

        # We bypass the sendChord() so volume isn't adjusted. Do it anyway!
        volume = trk.adjustVolume(volume, offset)

        # bypassing voice set, do it here
        trk.insertVoice(self.seq)

        # bypass even more
        trk.clearPending()

        # octave adjustment
        note += o.octave
        while note > 127:
            note -= 12
        while note < 0:
            note += 12
        # Add the noise.
        trk.sendNote(offset, o.duration, note, volume)


    def endVibration(self, stringNo, offset):
        """ kill the vibration on the string by sending out a note off """

        vibration = self._vibrating[stringNo]

        # first test if this string has been played before
        if vibration.note is None:
            return

        lastOffset = vibration.offset
        o = offset - self.artic[self.seq]   # move off point back before the next ON

        # ensure we turn this string OFF after the ON offset for this note
        while o+gbl.tickOffset <= lastOffset:
            o += 1

        gbl.mtrks[self.channel].addNoteOnToTrack(o, vibration.note, 0)  # v=0 ==note off

        # save the turn-off point for fret noise function
        self._lastOff[stringNo] = copy.copy(vibration) # this gets the note, volume
        self._lastOff[stringNo].offset = o + gbl.tickOffset # offset into track


        vibration.note = None
        vibration.offset = None


    def grooveFinish(self, offset):
        """ End all vibrations (ie output all outstanding note off). """

        for stringNo in range(len(self._vibrating)):
            self.endVibration(stringNo, offset)

    def fretboardNote(self, stringNo, chordList, startIdx, previousFret, chordBarreFretNo):
        """ Returns a single note that is is on the chord for one string
            Unlike a guitar string number the lowest string is string Number 0
        """

        openString = self._tuning[stringNo] + self._capoFretNo + chordBarreFretNo

        fretNotes = {}
        for n, note in enumerate(chordList[startIdx:]):
            fretNo = note - openString % 12
            while fretNo < 0:
                fretNo += 12
            fretNo %= 12

            fret = PlecStruct()
            fret.pitch = self.adjustNote(openString + fretNo)
            fret.fretNo = fretNo
            fret.chordIndex = n + startIdx
            fretNotes[fretNo] = fret

        # Try to guess where the next finger position in the chord goes
        # without trying to stretch the hand too much.
        if previousFret:
            if previousFret.fretNo in fretNotes:
                return fretNotes[previousFret.fretNo]
            if (previousFret.fretNo - 1) in fretNotes:
                return fretNotes[previousFret.fretNo - 1]
            if (previousFret.fretNo - 2) in fretNotes:
                return fretNotes[previousFret.fretNo - 2]
            if (previousFret.fretNo + 1) in fretNotes and previousFret.fretNo <= 3:
                return fretNotes[previousFret.fretNo + 1]

        lowest = min(fretNotes.keys())

        return fretNotes[lowest]

    def fretboardNotes(self, chordList, chordBarreFretNo):
        notes = []

        if len(chordList) >= 5 or len(self._tuning) >= 8:
            # I don't know how to handle five or more notes in a chord
            # or more than 8 strings (eg a harp)
            # so just do a single pass and see what happens
            previousFret = None
            for stringNo in range(len(self._tuning)):
                fret = self.fretboardNote(stringNo, chordList, 0, previousFret, chordBarreFretNo)
                previousFret = fret
                notes.append(fret)

        else:
            #find the triad chord (the first three notes of the chord) first
            previousFret = None
            for stringNo in range(len(self._tuning)):
                fret = self.fretboardNote(stringNo, chordList[:3], 0,
                                          previousFret, chordBarreFretNo)
                previousFret = fret
                notes.append(fret)

            if len(chordList) == 4:
                # Now put in missing 7th (or what ever it is called) but
                # start searching from the top string

                for stringNo in reversed(range(len(self._tuning))):

                    fret = self.fretboardNote(stringNo, chordList, 3, None, chordBarreFretNo)
                    if fret.fretNo <= 4:
                        notes[stringNo] = fret

                        # now go back up to the top string to make sure there is no bunching
                        stringNo += 1
                        while stringNo < len(self._tuning):
                            notes[stringNo] = self.fretboardNote(stringNo, chordList, 0,
                                                                 None, chordBarreFretNo)
                            stringNo += 1
                        break

        # look for and mark duplicates
        notes[0].duplicate = False
        for stringNo in range(1, len(self._tuning)):
            if notes[stringNo - 1].pitch == notes[stringNo].pitch:
                notes[stringNo].duplicate = True
            else:
                notes[stringNo].duplicate = False

        return notes

    def fretboardShapes(self, chordName, chordList):
        shape = self._chordShapes[chordName]
        notes = []
        stringNo = -1
        for fretNo in shape:
            stringNo += 1
            fret = PlecStruct()
            try:
                fretNo = int(fretNo)
            except:
                fretNo = None

            openString = self._tuning[stringNo]
            fret.fretNo = fretNo
            fret.chordIndex = stringNo
            fret.duplicate = False
            if fretNo == None:
                fret.pitch = openString
            else:
                fret.pitch = self.adjustNote(openString + fretNo)
            notes.append(fret)

        return notes

    def trackBar(self, pattern, ctable):
        """ Do a plectrum bar.

            Called from self.bar()

        """

        sc = self.seq
        
        for p in pattern:
            try:
                ct = self.getChordInPos(p.offset, ctable)
                chordName = ct.chord.name
                chordList = ct.chord.noteList   # catch no noteList attribute
            except AttributeError:
                continue

            if ct.plectrumZ:
                continue

            if len(self._tuning) != len(p.pluckVol):
                error("%s: Pattern and tuning lengths (%s, %s) do not match. "
                      "Was tuning changed?" % (self.name, len(p.pluckVol), len(self._tuning)))

            chordBarreFretNo = 0
            if ct.name.startswith('+'):
                chordBarreFretNo += 12
            if ct.name.startswith('-'):
                chordBarreFretNo += -12

            chordBarreFretNo += ct.chord.barre

            if mma.MMA.debug.debug or mma.MMA.debug.plecShow:
                self.printChordShape(ct, chordBarreFretNo)

            plectrumNoteOnList = []  # for debugging only

            # Find how many strings have been plucked this time
            pluckStringCount = 0
            for vol in p.pluckVol:
                if vol == -1:
                    continue
                pluckStringCount += 1

            pluckStringIndex = 0

            if chordName in self._chordShapes:
                notes = self.fretboardShapes(chordName, chordList)
            else:
                notes = self.fretboardNotes(chordList, chordBarreFretNo)

            fretNoiseCount = 0
            for stringNo, vol in enumerate(p.pluckVol):
                if p.strum:
                    if self.strumCenter == 0:
                        # start strum at end of beat
                        strumOffset = p.offset + p.strum * pluckStringIndex
                    elif self.strumCenter == 1:
                        # the centre of the strum is on the beat
                        strumOffset = p.offset + p.strum * \
                            (pluckStringIndex - pluckStringCount / 2.0)
                    else:
                        # start strum at end of beat
                        strumOffset = p.offset + p.strum * \
                            (pluckStringCount - pluckStringIndex - 1)

                    if p.strum < 0:
                        strumOffset += -p.strum * pluckStringCount
                else:
                    strumOffset = p.offset

                if vol == -1:
                    # silence this stringNo if the note on this string has changed
                    # even if this stringNo has not been plucked or muted
                    if notes[stringNo].pitch != self._vibrating[stringNo].note:
                        self.endVibration(stringNo, strumOffset)

                    continue   # this string has not been plucked or damped

                pluckStringIndex += 1
                
                self.endVibration(stringNo, strumOffset)
                if vol >= 1:

                    note = notes[stringNo].pitch

                    if notes[stringNo].duplicate:
                        if mma.MMA.debug.debug:
                            dPrint("%s: Ignoring duplicate note %d." % (self.name, note))
                        continue

                    if notes[stringNo].fretNo == None:
                        if mma.MMA.debug.debug:
                            dPrint("%s: Ignoring muted note %d." % (self.name, note))
                        continue

                    outputVolume = self.adjustVolume(vol, p.offset)

                    ontime = gbl.mtrks[self.channel].addNoteOnToTrack(strumOffset, note,
                                                             outputVolume,
                                                             self.rTime[sc][0],
                                                             self.rTime[sc][1])

                    self._vibrating[stringNo].note = note
                    self._vibrating[stringNo].offset = ontime  # save the start time
                    self._vibrating[stringNo].volume = outputVolume # and volume
                    if outputVolume == 0:
                        self._vibrating[stringNo].note = None

                    # Do fretnoise stuff if enabled
                    
                    if self.lastChord != ct.name and self.fretNoise  \
                       and outputVolume and stringNo in self.fretNoise.strings:
                        if fretNoiseCount < self.fretNoise.max:
                            self.doFretNoise(stringNo, note, strumOffset)
                            fretNoiseCount += 1

                    plectrumNoteOnList.append(note)  # for debugging only

                self.lastChord = ct.name

            if mma.MMA.debug.debug:
                dPrint("%s: channel=%s offset=%s chordList=%s NoteOn=%s." % 
                    (self.name, self.channel, p.offset + gbl.tickOffset,
                        chordList, plectrumNoteOnList))

    def printChordShape(self, chordTable, chordBarreFretNo=0):

        chordList = chordTable.chord.noteList

        # catch the case when there is no noteList attribute
        if hasattr(self, 'previousChordList'):
            if chordList == self.previousChordList and \
                    chordBarreFretNo == self.previousFretNo \
                    and not ( chordTable.name in self._chordShapes \
                              and self._chordShapes[chordTable.name] != self.previousShape ) :
                return
        self.previousChordList = chordList
        self.previousFretNo = chordBarreFretNo
        if  chordTable.name in self._chordShapes:
            self.previousShape = self._chordShapes[chordTable.name]
        else:
            self.previousShape = None

        capoFretNo = self._capoFretNo
        if chordTable.name in self._chordShapes:
            notes = self.fretboardShapes(chordTable.name, chordList)
            capoFretNo = 0
            chordBarreFretNo = 0
        else:
            notes = self.fretboardNotes(chordList, chordBarreFretNo)
        notes.reverse()

        printStart = 0
        startFretNo = capoFretNo + chordBarreFretNo
        if startFretNo < 0:
            printStart = startFretNo

        dPrint("%s %s chord %s" % (self.name, chordTable.name,  chordList))
        for stringNo, openNote in enumerate(reversed(self._tuning)):
            openNote = self.adjustNote(openNote)  # puts into middle octave 60==5*12
            note = notes[stringNo].pitch

            msg = ["%s   %3d" % (self.name, openNote + capoFretNo)]
            finger = note - openNote
            for fretNo in range(printStart, 20):
                if fretNo == 0 and notes[stringNo].fretNo == None and capoFretNo == 0:
                    msg.append("X")
                elif fretNo == 0 and capoFretNo == 0:
                    msg.append("|")
                elif fretNo == capoFretNo and chordBarreFretNo == 0:
                    if notes[stringNo].fretNo == None:
                        msg.append("X")
                    else:
                        msg.append("$")
                elif fretNo == finger:
                    msg.append("*")
                elif fretNo == capoFretNo:
                    msg.append("$")
                elif fretNo == startFretNo:
                    msg.append(":")
                elif fretNo == 0:
                    msg.append("|")
                else:
                    msg.append("-")

            msg.append("%d" % (notes[stringNo].chordIndex))
            if notes[stringNo].fretNo != None:
                msg.append("  %d" % (note))
            if notes[stringNo].duplicate:
                msg.append("  duplicate")

            dPrint(' '.join(msg))
        dPrint('\n')

def noteNameToMidiPitch(s):
    """ Convert a name ('e', 'g#') to a MIDI pitch. """

    tb = {'c': 0, 'c#': 1, 'd&': 1, 'd': 2, 'd#': 3, 'e&': 3,
          'e': 4, 'f&': 4, 'e#': 5, 'f': 5, 'f#': 6, 'g&': 6,
          'g': 7, 'g#': 8, 'a&': 8, 'a': 9, 'a#': 10, 'b&': 10,
          'b': 11, 'b&': 11, 'c&': 11, 'b#': 0}

    # strip and count trailing '+' and '-'

    if '-' in s and '+' in s:
        return None

    adjust = 0
    while s.endswith('-'):
        adjust -= 12
        s = s[:-1]
    while s.endswith('+'):
        adjust += 12
        s = s[:-1]

    try:
        value = tb[s]  # puts into middle octave 60==5*12
    except:
        return None

    return value + adjust


#not used. Leave for debugging??
def MidiPitch2NoteName(value):
    nameLookUp = ['c', 'c#', 'd', 'e&', 'e', 'f', 'f#', 'g', 'g#', 'a', 'b&', 'b']




# macros.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

The macros are stored, set and parsed in this single-instance
class. At the top of MMAparse an instance in created with
something like:     macros=MMMmacros.Macros().
"""

import random
import datetime

import mma.MMA.midiC
import mma.MMA.translate
import mma.MMA.volume
import mma.MMA.grooves
import mma.MMA.parse
import mma.MMA.parseCL
import mma.MMA.player
import mma.MMA.seqrnd
import mma.MMA.midinote
import mma.MMA.swing
import mma.MMA.ornament
import mma.MMA.rpitch
import mma.MMA.chords
import mma.MMA.debug

from . import gbl

from   mma.MMA.notelen import getNoteLen
from   mma.MMA.keysig import keySig
from   mma.MMA.timesig import timeSig
from   mma.MMA.lyric import lyric
from   mma.MMA.common import *
from   mma.MMA.safe_eval import safe_eval

def sliceVariable(p, sl):
    """ Slice a variable. Used by macro expand. """

    try:
        new = eval('p' + "[" + sl + "]")
    except IndexError:
        error("Index '%s' out of range." % sl)
    except:
        error("Index error in '%s'." % sl)

    if ":" not in sl:
        new = [new]

    return new


class Macros:
    vars = {}            # storage
    expandMode = 1        # flag for variable expansion
    pushstack = []

    def __init__(self):

        self.vars = {}

    def clear(self, ln):
        if ln:
            error("VarClear does not take an argument.")
        self.vars = {}
        if mma.MMA.debug.debug:
            dPrint("All variable definitions cleared.")

    def stackValue(self, s):
        self.pushstack.append(' '.join(s))

    def sysvar(self, s):
        """ Create an internal macro. """
        
        # Simple/global     system values

        if s == 'CHORDADJUST':
            return ' '.join([ "%s=%s" % (a, mma.MMA.chords.cdAdjust[a]) 
                              for a in sorted(mma.MMA.chords.cdAdjust)])
        elif s == 'FILENAME':
            return str(gbl.inpath.fname)

        elif s == 'KEYSIG':
            return keySig.getKeysig()

        elif s == 'TIME':
            return str(gbl.QperBar)

        elif s == 'CTABS':
            return ','.join([ str((float(x) / gbl.BperQ) + 1) for x in mma.MMA.parseCL.chordTabs])

        elif s == 'TIMESIG':
            return timeSig.getAscii()

        elif s == 'TEMPO':
            return str(gbl.tempo)

        elif s == 'OFFSET':
            return str(gbl.tickOffset)
        
        elif s == 'SONGFILENAME':
            return str(gbl.infile)

        elif s == 'VOLUME':
            return str(int(mma.MMA.volume.volume * 100))  # INT() is important

        elif s == 'VOLUMERATIO':
            return str((mma.MMA.volume.vTRatio * 100))

        elif s == 'LASTVOLUME':
            return str(int(mma.MMA.volume.lastVolume * 100))

        elif s == 'GROOVE':
            return mma.MMA.grooves.currentGroove

        elif s == 'GROOVELIST':
            return ' '.join(sorted([x for x in mma.MMA.grooves.glist.keys() if isinstance(x, str)]))

        elif s == 'TRACKLIST':
            return ' '.join(sorted(gbl.tnames.keys()))

        elif s == 'LASTGROOVE':
            return mma.MMA.grooves.lastGroove

        elif s == 'PLUGINS':
            from mma.MMA.regplug import simplePlugs  # to avoid circular import error
            return ' '.join(simplePlugs)

        elif s == 'TRACKPLUGINS':
            from mma.MMA.regplug import trackPlugs  # to avoid circular import error
            return ' '.join(trackPlugs)
         
        elif s == 'DATAPLUGINS':
            from mma.MMA.regplug import dataPlugs  # to avoid circular import error
            return ' '.join(dataPlugs)
        
        elif s == 'SEQ':
            return str(gbl.seqCount)

        elif s == 'SEQRND':
            if mma.MMA.seqrnd.seqRnd[0] == 0:
                return "Off"
            if mma.MMA.seqrnd.seqRnd[0] == 1:
                return "On"
            return ' '.join(mma.MMA.seqrnd.seqRnd[1:])

        elif s == 'SEQSIZE':
            return str(gbl.seqSize)

        elif s == 'SWINGMODE':
            return mma.MMA.swing.settings()

        elif s == 'TICKPOS':
            return str(gbl.tickOffset)
        
        elif s == 'TRANSPOSE':
            return str(gbl.transpose)

        elif s == 'STACKVALUE':
            if not self.pushstack:
                error("Empty push/pull variable stack")
            return self.pushstack.pop()

        elif s == 'DEBUG':
            return mma.MMA.debug.getFlags()
        
        elif s == 'LASTDEBUG':
            return mma.MMA.debug.getLFlags()
            
        elif s == 'VEXPAND':
            if self.expandMode:
                return "On"
            else:
                return "Off"

        elif s == "MIDIPLAYER":
            return "%s Background=%s Delay=%s." % \
                (' '.join(mma.MMA.player.midiPlayer), mma.MMA.player.inBackGround,
                 mma.MMA.player.waitTime)

        elif s == "MIDISPLIT":
            return ' '.join([str(x) for x in mma.MMA.midi.splitChannels])

        elif s.startswith("NOTELEN(") and s.endswith(")"):
            return "%sT" % getNoteLen(s[8:-1])
            
        elif s == 'SEQRNDWEIGHT':
            return ' '.join([str(x) for x in mma.MMA.seqrnd.seqRndWeight])

        elif s == 'AUTOLIBPATH':
            return ' '.join(mma.MMA.paths.libDirs)

        elif s == 'LIBPATH':
            return ' '.join(mma.MMA.paths.libPath)

        elif s == 'MMAPATH':
            return gbl.MMAdir

        elif s == 'INCPATH':
            return ' '.join(mma.MMA.paths.incPath)

        elif s == 'VOICETR':
            return mma.MMA.translate.vtable.retlist()

        elif s == 'TONETR':
            return mma.MMA.translate.dtable.retlist()

        elif s == 'OUTPATH':
            return gbl.outPath

        elif s == 'BARNUM':
            return str(gbl.barNum + 1)

        elif s == 'LINENUM':
            return str(gbl.lineno)

        elif s == 'LYRIC':
            return lyric.setting()

        # Some time/date macros. Useful for generating copyright strings

        elif s == 'DATEYEAR':
            return str(datetime.datetime.now().year)

        elif s == 'DATEDATE':
            return datetime.datetime.now().strftime("%Y-%m-%d")

        elif s == 'DATETIME':
            return datetime.datetime.now().strftime("%H:%M:%S")

        
        # Track vars ... these are in format TRACKNAME_VAR

        a = s.rfind('_')
        if a == -1:
            error("Unknown system variable $_%s" % s)

        tname = s[:a]
        func = s[a+1:]

        try:
            t = gbl.tnames[tname]
        except KeyError:
            error("System variable $_%s refers to nonexistent track." % s)

        if func == 'ACCENT':
            r = []
            for s in t.accent:
                r.append("{")
                for b, v in s:
                    r.append('%g' % (b/float(gbl.BperQ)+1))
                    r.append(str(int(v * 100)))
                r.append("}")
            return ' '.join(r)

        elif func == 'ARTICULATE':
            return ' '.join([str(x) for x in t.artic])

        elif func == 'CHORDS':
            r = []
            for l in t.chord:
                r.append('{' + ' '.join(l) + '}')
            return ' '.join(r)

        elif func == 'CHANNEL':
            return str(t.channel)

        elif func == 'COMPRESS':
            return ' '.join([str(x) for x in t.compress])

        elif func == 'DELAY':
            return ' '.join([str(x) for x in t.delay])

        elif func == 'DIRECTION':
            if t.vtype == 'ARIA':
                return ' '.join([str(x) for x in t.selectDir])
            else:
                return ' '.join([str(x) for x in t.direction])

        elif func == 'DUPROOT':
            if t.vtype != "CHORD":
                error("Only CHORD tracks have DUPROOT")
            return t.getDupRootSetting()

        elif func == 'FRETNOISE':
            return t.getFretNoiseOptions()

        elif func == 'HARMONY':
            return ' '.join([str(x) for x in t.harmony])

        elif func == 'HARMONYONLY':
            return ' '.join([str(x) for x in t.harmonyOnly])

        elif func == 'HARMONYVOLUME':
            return ' '.join([str(int(i * 100)) for i in t.harmonyVolume])

        elif func == 'INVERT':
            return ' '.join([str(x) for x in t.invert])

        elif func == 'LIMIT':
            return str(t.chordLimit)

        elif func == 'MALLET':
            if t.vtype not in ("SOLO", "MELODY"):
                error("Mallet only valid in SOLO and MELODY tracks")
            return "Mallet Rate=%i Decay=%i" % (t.mallet, t.malletDecay*100)

        elif func == 'MIDINOTE':
            return mma.MMA.midinote.mopts(t)

        elif func == 'MIDIVOLUME':
            return "%s" % t.cVolume

        elif func == 'OCTAVE':
            return ' '.join([str(i//12) for i in t.octave])

        elif func == 'MOCTAVE':
            return ' '.join([str((i//12)-1) for i in t.octave])

        elif func == 'ORNAMENT':
            return mma.MMA.ornament.getOrnOpts(t)
        
        elif func == 'PLUGINS':
            from mma.MMA.regplug import trackPlugs  # avoids circular import
            return ' '.join(trackPlugs)
        
        elif func == 'RANGE':
            return ' '.join([str(x) for x in t.chordRange])

        elif func == 'RSKIP':
            m = ''
            if t.rSkipBeats:
                m = "Beats=%s " % ','.join(['%g' % (x/float(gbl.BperQ)+1) for x in t.rSkipBeats])
            m += ' '.join([str(int(i * 100)) for i in t.rSkip])
            return m

        elif func == 'RDURATION':
            tmp = []
            for a1, a2 in t.rDuration:
                a1 = int(a1 * 100)
                a2 = int(a2 * 100)
                if a1 == a2:
                    tmp.append('%s' % abs(a1))
                else:
                    tmp.append('%s,%s' % (a1, a2))

            return ' '. join(tmp)

        elif func == 'RTIME':
            tmp = []
            for a1, a2 in t.rTime:
                if a1 == a2:
                    tmp.append('%s' % abs(a1))
                else:
                    tmp.append('%s,%s' % (a1, a2))
            return ' '.join(tmp)

        elif func == 'RVOLUME':
            tmp = []
            for a1, a2 in t.rVolume:
                a1 = int(a1 * 100)
                a2 = int(a2 * 100)
                if a1 == a2:
                    tmp.append('%s' % abs(a1))
                else:
                    tmp.append('%s,%s' % (a1, a2))
            return ' '.join(tmp)

        elif func == 'RPITCH':
            return mma.MMA.rpitch.getOpts(t)
            
                    
        elif func == 'SEQUENCE':
            tmp = []
            for a in range(gbl.seqSize):
                tmp.append('{' + t.formatPattern(t.sequence[a]) + '}')
            return ' '.join(tmp)

        elif func == 'SEQRND':
            if t.seqRnd:
                return 'On'
            else:
                return 'Off'

        elif func == 'SEQRNDWEIGHT':
            return ' '.join([str(x) for x in t.seqRndWeight])

        elif func == 'SPAN':
            return "%s %s" % (t.spanStart, t.spanEnd)

        elif func == 'STICKY':
            if t.sticky:
                return "True"
            else:
                return "False"
            

        elif func == 'STRUM':
            r = []
            for v in t.strum:
                if v is None:
                    r.append("0")
                else:
                    a, b = v
                    if a == b:
                        r.append("%s" % a)
                    else:
                        r.append("%s,%s" % (a, b))

            return ' '.join(r)

        elif func == 'STRUMADD':
            return ' '.join([str(x) for x in t.strumAdd])

        elif func == 'TRIGGER':
            return mma.MMA.trigger.getTriggerOptions(t)

        elif func == 'TONE':
            if t.vtype in ('MELODY', 'SOLO'):
                if not t.drumType:
                    error("Melody/Solo tracks must be DRUMTYPE for tone.")
                return str(mma.MMA.midiC.valueToDrum(t.drumTone))

            elif t.vtype != 'DRUM':
                error("Tracktype %s doesn't have TONE" % t.vtype)

            return ' '.join([mma.MMA.midiC.valueToDrum(a) for a in t.toneList])

        elif func == 'UNIFY':
            return ' '.join([str(x) for x in t.unify])

        elif func == 'VOICE':
            return ' '.join([mma.MMA.midiC.valueToInst(a) for a in t.voice])

        elif func == 'VOICING':
            if t.vtype != 'CHORD':
                error("Only CHORD tracks have VOICING")
            t = t.voicing
            return "Mode=%s Range=%s Center=%s RMove=%s Move=%s Dir=%s" % \
                (t.mode, t.range, t.center, t.random, t.bcount, t.dir)

        elif func == 'VOLUME':
            return ' '.join([str(int(a * 100)) for a in t.volume])

        else:
            error("Unknown system track variable %s" % s)

    def expand(self, l):
        """ Loop though input line and make variable subsitutions.
            MMA variables are pretty simple ... any word starting
            with a "$xxx" is a variable.

            l - list

            RETURNS: new list with all subs done.
        """

        if not self.expandMode:
            return l

        gotmath = 0
        sliceVar = None

        while 1:          # Loop until no more subsitutions have been done
            sub = 0

            for i, s in enumerate(l):
                  if s[0] == '$':
                    
                    s = s[1:].upper()
                    if not s:
                        error("Illegal macro name '%s'." % l[i])

                    frst = s[0]  # first char after the leading '$'

                    if frst == '$':  # $$ - don't expand (done in IF clause)
                        continue

                    if frst == '(':   # flag math macro
                        gotmath = 1
                        continue

                    # pull slice notation off the end of the name

                    if s.endswith(']'):
                        x = s.rfind('[')
                        if not x:
                            error("No matching for '[' for trailing ']' in variable '%s'." % s)
                        sliceVar = s[x+1:-1]
                        s = s[:x]

                        """ Since we be using an 'eval' to do the actual slicing,
                            we check the slice string to make sure it's looking
                            valid. The easy way out makes as much sense as anything
                            else ... just step through the slice string and make
                            sure we ONLY have integers or empty slots.
                        """

                        for test in sliceVar.split(":"):
                            try:
                                test == '' or int(test)
                            except:
                                error("Invalid index in slice notation '%s'." % sliceVar)

                    else:
                        sliceVar = None

                    # we have a var, see if system or user. Set 'ex'

                    if frst == '_':   # system var
                        ex = self.sysvar(s[1:])

                    elif s in self.vars:  # user var?
                        ex = self.vars[s]

                    else:                 # unknown var...error
                        error("User variable '%s'  has not been defined" % s)

                    if isinstance(ex, list):  # MSET variable
                        if sliceVar:
                            ex = sliceVariable(ex, sliceVar)
                            sliceVar = None

                        if len(ex):
                            gbl.inpath.push(ex[1:], [gbl.lineno] * len(ex[1:]))
                            if len(ex):
                                ex = ex[0]
                            else:
                                ex = []

                    else:                       # regular SET variable
                        ex = ex.split()
                        if sliceVar:
                            ex = sliceVariable(ex, sliceVar)
                            sliceVar = None

                    """ we have a simple variable (ie $_TEMPO) converted to a list,
                        or a list-like var (ie $_Bass_Volume) converted to a list,
                        or the 1st element of a multi-line variable
                        We concat this into the existing line, process some more
                    """

                    l = l[:i] + ex + l[i+1:]

                    sub = 1
                    break

            if not sub:
                break

        # all the mma internal and system macros are expanded. Now check for math.

        if gotmath:
            l = ' '.join(l)   # join back into one line

            while 1:
                try:
                    s1 = l.index('$(')  # any '$(' left?
                except:
                    break               # nope, done
                # find trailing )
                nest = 0
                s2 = s1+2
                max = len(l)
                while 1:
                    if l[s2] == '(':
                        nest += 1
                    if l[s2] == ')':
                        if not nest:
                            break
                        else:
                            nest -= 1
                    s2 += 1
                    if s2 >= max:
                        error("Unmatched delimiter in '%s'." % l)

                l = l[:s1] + str(safe_eval(l[s1+2:s2].strip())) + l[s2+1:]

            l = l.split()

        return l

    def showvars(self, ln):
        """ Display all currently defined user variables. """

        if len(ln):
            for a in ln:
                a = a.upper()
                if a in self.vars:
                    print("$%s: %s" % (a, self.vars[a]))
                else:
                    print("$%s - not defined" % a)

        else:

            print("User variables defined:")
            kys = self.vars.keys()
            kys.sort()

            mx = 0

            for a in kys:                    # get longest name
                if len(a) > mx:
                    mx = len(a)

            mx = mx + 2

            for a in kys:
                print("     %-*s  %s" % (mx, '$'+a, self.vars[a]))

    def getvname(self, v):
        """ Helper routine to validate variable name. """

        if v[0] in ('$', '_'):
            error("Variable names cannot start with a '$' or '_'")
        if '[' in v or ']' in v:
            error("Variable names cannot contain [ or ] characters.")

        return v.upper()

    def rndvar(self, ln):
        """ Set a variable randomly from a list. """

        if len(ln) < 2:
            error("Use: RndSet Variable_Name <list of possible values>")

        v = self.getvname(ln[0])

        self.vars[v] = random.choice(ln[1:])

        if mma.MMA.debug.debug:
            dPrint("Variable $%s randomly set to '%s'" % (v, self.vars[v]))

    def newsetvar(self, ln):
        """ Set a new variable. Ignore if already set. """

        if not len(ln):
            error("Use: NSET VARIABLE_NAME [Value] [[+] [Value]]")

        if self.getvname(ln[0]) in self.vars:
            return

        self.setvar(ln)

    def setvar(self, ln):
        """ Set a variable. Note the difference between the next 2 lines:
                Set Bar BAR
                Set Foo AAA BBB $bar
                   $Foo == "AAA BBB BAR"
                Set Foo AAA + BBB + $bar
                   $Foo == "AAABBBBAR"

            The "+"s just strip out intervening spaces.
        """

        if len(ln) < 1:
            error("Use: SET VARIABLE_NAME [Value] [[+] [Value]]")

        v = self.getvname(ln.pop(0))

        t = ''
        addSpace = 0
        for i, a in enumerate(ln):
            if a == '+':
                addSpace = 0
                continue
            else:
                if addSpace:
                    t += ' '
                t += a
                addSpace = 1

        self.vars[v] = t

        if mma.MMA.debug.debug:
            dPrint("Variable $%s == '%s'" % (v, self.vars[v]))

    def msetvar(self, ln):
        """ Set a variable to a number of lines. """

        if len(ln) != 1:
            error("Use: MSET VARIABLE_NAME <lines> MsetEnd")

        v = self.getvname(ln[0])

        lm = []

        while 1:
            l = gbl.inpath.read()
            if not l:
                error("Reached EOF while looking for MSetEnd")
            cmd = l[0].upper()
            if cmd in ("MSETEND", 'ENDMSET'):
                if len(l) > 1:
                    error("No arguments permitted for MSetEnd/EndMSet")
                else:
                    break
            lm.append(l)

        self.vars[v] = lm
        
    def unsetvar(self, ln):
        """ Delete a variable reference. """

        if len(ln) != 1:
            error("Use: UNSET Variable")
        v = ln[0].upper()
        if v[0] == '_':
            error("Internal variables cannot be deleted or modified")

        if v in self.vars:
            del(macros.vars[v])

            if mma.MMA.debug.debug:
                dPrint("Variable '%s' UNSET" % v)
        else:
            warning("Attempt to UNSET nonexistent variable '%s'" % v)

    def vexpand(self, ln):

        if len(ln) == 1:
            cmd = ln[0].upper()
        else:
            cmd = ''

        if cmd == 'ON':
            self.expandMode = 1
            if mma.MMA.debug.debug:
                dPrint("Variable expansion ON")

        elif cmd == 'OFF':
            self.expandMode = 0
            if mma.MMA.debug.debug:
                dPrint("Variable expansion OFF")

        else:
            error("Use: Vexpand ON/Off")

    def varinc(self, ln):
        """ Increment  a variable. """

        if len(ln) == 1:
            inc = 1

        elif len(ln) == 2:
            inc = stof(ln[1], "Expecting a value (not %s) for Inc" % ln[1])

        else:
            error("Usage: INC Variable [value]")

        v = ln[0].upper()

        if v[0] == '_':
            error("Internal variables cannot be modified")

        if not v in self.vars:
            error("Variable '%s' not defined" % v)

        vl = stof(self.vars[v], "Variable must be a value to increment") + inc

        # lot of mma commands expect ints, so convert floats like 123.0 to 123

        if vl == int(vl):
            vl = int(vl)

        self.vars[v] = str(vl)

        if mma.MMA.debug.debug:
            dPrint("Variable '%s' INC to %s" % (v, self.vars[v]))

    def vardec(self, ln):
        """ Decrement a varaiable. """

        if len(ln) == 1:
            dec = 1

        elif len(ln) == 2:
            dec = stof(ln[1], "Expecting a value (not %s) for Inc" % ln[1])

        else:
            error("Usage: DEC Variable [value]")

        v = ln[0].upper()
        if v[0] == '_':
            error("Internal variables cannot be modified")

        if not v in self.vars:
            error("Variable '%s' not defined" % v)

        vl = stof(self.vars[v], "Variable must be a value to decrement") - dec

        # lot of mma commands expect ints, so convert floats like 123.0 to 123

        if vl == int(vl):
            vl = int(vl)

        self.vars[v] = str(vl)

        if mma.MMA.debug.debug:
            dPrint("Variable '%s' DEC to %s" % (v, self.vars[v]))

    def varIF(self, ln):
        """ Conditional variable if/then. """

        def expandV(l):
            """ Private func. """

            l = l.upper()

            if l[:2] == '$$':
                l = l.upper()
                l = l[2:]
                if not l in self.vars:
                    error("String Variable '%s' does not exist" % l)
                l = self.vars[l]

            try:
                v = float(l)
            except:
                try:
                    v = int(l, 0)  # this lets us convert HEX/OCTAL
                except:
                    v = None

            return(l.upper(), v)

        def readblk():
            """ Private, reads a block until ENDIF, IFEND or ELSE.
                Return (Terminator, lines[], linenumbers[] )
            """

            q = []
            qnum = []
            nesting = 0

            while 1:
                l = gbl.inpath.read()
                if not l:
                    error("EOF reached while looking for EndIf")

                cmd = l[0].upper()
                if cmd == 'IF':
                    nesting += 1
                if cmd in ("IFEND", 'ENDIF', 'ELSE'):
                    if len(l) > 1:
                        error("No arguments permitted for IfEnd/EndIf/Else")
                    if not nesting:
                        break
                    if cmd != 'ELSE':
                        nesting -= 1

                q.append(l)
                qnum.append(gbl.lineno)

            return (cmd, q, qnum)

        if len(ln) < 2:
            error("Usage: IF <Operator> ")

        action = ln[0].upper()

        # 1. do the unary options: DEF, NDEF

        if action in ('DEF', 'NDEF'):
            if len(ln) != 2:
                error("Usage: IF %s VariableName" % action)

            v = ln[1].upper()

            if action == 'DEF':
                compare = v in self.vars
            elif action == 'NDEF':
                compare = (v not in self.vars)
            else:
                error("Unreachable unary conditional")  # can't get here

        # 2. Binary ops: EQ, NE, etc.

        elif action in ('LT', '<', 'LE', '<=', 'EQ', '==', 'GE', '>=', 'GT', '>', 'NE', '!='):
            if len(ln) != 3:
                error("Usage: VARS %s Value1 Value2" % action)

            s1, v1 = expandV(ln[1])
            s2, v2 = expandV(ln[2])

            # Make the comparison to strings or values. If either arg
            # is NOT a value, use string values for both.
            if None in (v1, v2):
                v1, v2 = s1, s2

            if action == 'LT' or action == '<':
                compare = (v1 < v2)
            elif action == 'LE' or action == '<=':
                compare = (v1 <= v2)
            elif action == 'EQ' or action == '==':
                compare = (v1 == v2)
            elif action == 'GE' or action == '>=':
                compare = (v1 >= v2)
            elif action == 'GT' or action == '>':
                compare = (v1 > v2)
            elif action == 'NE' or action == '!=':
                compare = (v1 != v2)
            else:
                error("Unreachable binary conditional")  # can't get here

        else:
            error("Usage: IF <CONDITON> ...")

        """ Go read until end of if block.
            We shove the block back if the compare was true.
            Unless, the block is terminated by an ELSE ... then we need
            to read another block and push back one of the two.
        """

        cmd, q, qnum = readblk()

        if cmd == 'ELSE':
            cmd, q1, qnum1 = readblk()

            if cmd == 'ELSE':
                error("Only one ELSE is permitted in IF construct")

            if not compare:
                compare = 1
                q = q1
                qnum = qnum1

        if compare:
            gbl.inpath.push(q, qnum)


macros = Macros()
